import logging
import os
import pathlib

import pandas as pd
from ruamel import yaml

from cellpy.exceptions import FileNotFound
from cellpy.parameters import prms
from cellpy.parameters.internal_settings import ATTRS_CELLPYFILE, get_headers_normal
from cellpy.readers.core import Cell, FileID, check64bit, humanize_bytes
from cellpy.readers.instruments.base import Loader

DEFAULT_CONFIG = {
    "structure": {
        "format": "csv",
        "table_name": None,
        "header_definitions": "labels",
        "comment_chars": ("#", "!"),
        "sep": ";",
        "locate_start_data_by": "line_number",
        "locate_end_data_by": "EOF",
        "locate_vars_by": "key_value_pairs",
        "start_data": 19,
        "header_info_line": 1,
        "start_data_offset": 0,
        "header_info_parse": "in_key",
        "header_info_splitter": ";",
        "file_type_id_line": 0,
        "file_type_id_match": None,
    },
    "variables": {
        "mass": "mass",
        "total_mass": "total_mass",
        "schedule_file": "schedule_file",
        "schedule": "schedule",
        "creator": "operator",
        "loaded_from": "loaded_from",
        "channel_index": "channel_index",
        "channel_number": "channel_number",
        "item_ID": "instrument",
        "test_ID": "test_name",
        "cell_name": "cell",
        "material": "material",
        "counter_electrode": "counter",
        "reference_electrode": "reference",
        "start_datetime": "date",
        "fid_last_modification_time": "last_modified",
        "fid_size": "size",
        "fid_last_accessed": "last_accessed",
    },
    "headers": {
        "data_point_txt": "index",
        "charge_capacity_txt": "charge_capacity",
        "current_txt": "current",
        "cycle_index_txt": "cycle",
        "datetime_txt": "date_stamp",
        "discharge_capacity_txt": "discharge_Capacity",
        "step_index_txt": "step",
        "step_time_txt": "step_time",
        "test_time_txt": "test_time",
        "voltage_txt": "voltage",
    },
    "units": {"current": 0.001, "charge": 0.001, "mass": 0.001, "specific": 1.0},
    "limits": {
        "current_hard": 0.0000000000001,
        "current_soft": 0.00001,
        "stable_current_hard": 2.0,
        "stable_current_soft": 4.0,
        "stable_voltage_hard": 2.0,
        "stable_voltage_soft": 4.0,
        "stable_charge_hard": 0.9,
        "stable_charge_soft": 5.0,
        "ir_change": 0.00001,
    },
}


class CustomLoader(Loader):
    """Class for loading cell data from custom formatted files.

    The file that contains the description of the custom data file
    should be given by issuing the
    pick_definition_file or given in the config file
    (prms.Instruments.custom_instrument_definitions_file)

    The format of the custom data file should be on the form

    ...
        # comment
        # ...
        variable sep value
        variable sep value
        ...
        header1 sep header2 sep ...
        value1  sep value2  sep ...
        ...

    where sep is either defined in the description file or the
    config file.

    The definition file should use the YAML format and it
    must contain

    xxx
    xxx

    [doc not finished yet]
    """

    name = "old_custom"
    raw_ext = "*"

    def __init__(self, *args, **kwargs):
        """initiates the class"""

        self.logger = logging.getLogger(__name__)
        self.headers_normal = get_headers_normal()
        self.definition_file = self._pick_definition_file()
        self.units = None
        self.limits = None
        self.headers = None
        self.variables = None
        self.structure = None
        self._parse_definition_file()

    @staticmethod
    def _pick_definition_file():
        logging.debug(
            f"instrument file: {prms.Instruments.custom_instrument_definitions_file}"
        )
        return prms.Instruments.custom_instrument_definitions_file

    @staticmethod
    def _load_definition_file():
        definitions_file = pathlib.Path(
            prms.Instruments.custom_instrument_definitions_file
        )
        if not definitions_file.is_file():
            raise FileNotFound(
                f"Custom definitions file not found ({definitions_file})"
            )
        yml = yaml.YAML()
        with open(definitions_file, "r") as ff:
            settings = yml.load(ff.read())
        return settings

    # TODO: @jepe - create yaml file example (from DEFAULT_CONFIG)

    def _parse_definition_file(self):
        if self.definition_file is None:
            logging.info("no definition file for custom format")
            logging.info("using default settings")
            settings = DEFAULT_CONFIG
        else:
            logging.info("loading definition file for custom format")
            settings = self._load_definition_file()

        self.units = settings["units"]
        self.limits = settings["limits"]
        self.headers = settings["headers"]
        self.variables = settings["variables"]
        self.structure = settings["structure"]

    def get_raw_units(self):
        return self.units

    def get_raw_limits(self):
        return self.limits

    def _find_data_start(self, file_name, sep):
        if self.structure["locate_start_data_by"] != "line_number":
            raise NotImplementedError
        if not self.structure["start_data"] is None:
            return self.structure["start_data"] + self.structure["start_data_offset"]

        else:
            logging.debug("searching for line where data starts")
            header_info_line = self.structure["header_info_line"]
            header_info_parse = self.structure["header_info_parse"]
            header_info_splitter = self.structure["header_info_splitter"]
            header_info_line = self.structure["header_info_line"]

            with open(file_name, "rb") as fp:
                for i, line_ in enumerate(fp):
                    if i == header_info_line:
                        line = line_.strip()
                        line = line.decode()
                        break

            if header_info_parse == "in_key":
                _, v = line.split(header_info_splitter)
            else:
                _, v = line.split(sep)
            v = int(v)
            return v

    def loader(self, file_name, **kwargs):
        """Custom loader method

        Args:
            file_name (path): name of the file with the raw data
            **kwargs: optional key-word arguments
                pre_processor_hook (callable): function to apply before returning
                    the data (run after renaming the cols but before changing the structure)

        Returns:
            List of cellpy Cell objects.
        """
        pre_processor_hook = kwargs.pop("pre_processor_hook", None)
        new_tests = []
        var_dict = dict()

        cycle_index_hdr = self.headers_normal["cycle_index_txt"]
        charge_cap_hdr = self.headers_normal["charge_capacity_txt"]
        discharge_cap_hdr = self.headers_normal["discharge_capacity_txt"]
        datetime_hdr = self.headers_normal["datetime_txt"]
        data_point = self.headers_normal["data_point_txt"]
        step_time_hdr = self.headers_normal["step_time_txt"]
        test_time_hdr = self.headers_normal["test_time_txt"]

        if not os.path.isfile(file_name):
            self.logger.info("Missing file_\n   %s" % file_name)
            return

        # find out strategy (based on structure)
        if self.structure["format"] not in ["csv", "xlsx", "xls"]:
            print(self.structure["format"])
            raise NotImplementedError
        else:
            logging.debug(self.structure["format"])

        if self.structure["format"] == "csv":
            sep = self.structure.get("sep", prms.Reader.sep)

            locate_vars_by = self.structure.get("locate_vars_by", "key_value_pairs")
            comment_chars = self.structure.get("comment_chars", ["#", "!"])
            header_row = self.structure.get("start_data", None)
            if header_row is None:
                header_row = self._find_data_start(file_name, sep)

            # parsing the top part of the file, looking for variables
            var_lines = []
            with open(file_name, "rb") as fp:
                for i, line in enumerate(fp):
                    if i < header_row:
                        line = line.strip()
                        try:
                            line = line.decode()
                        except UnicodeDecodeError:
                            logging.debug(
                                "UnicodeDecodeError: " "skipping this line: " f"{line}"
                            )
                        else:
                            if line.startswith(comment_chars):
                                logging.debug(f"Comment: {line}")
                            else:
                                var_lines.append(line)
                    else:
                        break

            if locate_vars_by == "key_value_pairs":
                for line in var_lines:
                    parts = line.split(sep)
                    try:
                        var_dict[parts[0]] = parts[1]
                    except IndexError as e:
                        logging.debug(f"{e}\ncould not split var-value\n{line}")

            else:
                raise NotImplementedError

        data = Cell()
        data.loaded_from = file_name
        fid = self._generate_fid(file_name, var_dict)

        # parsing cellpydata attributes
        for attribute in ATTRS_CELLPYFILE:
            key = self.variables.get(attribute, None)
            val = var_dict.pop(key, None)
            # print(f"{attribute} -> {key}:{val}")
            if val:
                if key in ["mass"]:
                    val = float(val)
                # print(f"mass, {attribute}: {val}")
                setattr(data, attribute, val)

        data.raw_data_files.append(fid)

        # setting optional attributes (will be implemented later I hope)
        key = self.variables.get("total_mass", None)
        if key:
            total_mass = var_dict.pop(key, None)
            logging.debug("total_mass is given, but not propagated")

        logging.debug(f"unused vars: {var_dict}")
        if self.structure["format"] == "csv":
            raw = self._parse_csv_data(file_name, sep, header_row)

        elif self.structure["format"] == "xlsx":
            raw = self._parse_xlsx_data(file_name)

        elif self.structure["format"] == "xls":
            raw = self._parse_xls_data(file_name)

        raw = self._rename_cols(raw)
        if pre_processor_hook is not None:
            logging.debug("running pre-processing-hook")
            raw = pre_processor_hook(raw)

        capacity_structure = self.structure.get("capacity_structure", "cellpy")
        if capacity_structure == "cellpy":
            logging.debug(
                "standard cellpy structure - no additional processing of capacity columns needed"
            )
        elif capacity_structure == "one_col_state":
            # TODO: make this a function or method
            # TODO: currently, the user needs to assign the name of the capacity column to the charge_capacity
            #  variable in the yaml file. Should improve this later.
            delimiter = "::"
            cap_col = "charge_capacity"
            col = self.structure["state_col"]
            charge_key = self.structure["charge_key"]
            charge_key = charge_key.split(delimiter)
            discharge_key = self.structure["discharge_key"]
            discharge_key = discharge_key.split(delimiter)

            raw["new_c"] = 0
            raw["new_d"] = 0

            cycle_numbers = raw[cycle_index_hdr].unique()
            # cell_type = prms.Reader.cycle_mode
            good_cycles = []
            bad_cycles = []
            for i in cycle_numbers:
                try:
                    charge_cap = raw.loc[
                        (raw[col].isin(charge_key)) & (raw[cycle_index_hdr] == i),
                        [data_point, cap_col],
                    ]
                    discharge_cap = raw.loc[
                        (raw[col].isin(discharge_key)) & (raw[cycle_index_hdr] == i),
                        [data_point, cap_col],
                    ]

                    if not charge_cap.empty:
                        charge_cap_last_index, charge_cap_last_val = charge_cap.iloc[-1]
                        raw["new_c"].update(charge_cap[cap_col])

                        raw.loc[
                            (raw[data_point] > charge_cap_last_index)
                            & (raw[cycle_index_hdr] == i),
                            "new_c",
                        ] = charge_cap_last_val

                    if not discharge_cap.empty:
                        (
                            discharge_cap_last_index,
                            discharge_cap_last_val,
                        ) = discharge_cap.iloc[-1]
                        raw["new_d"].update(discharge_cap[cap_col])

                        raw.loc[
                            (raw[data_point] > discharge_cap_last_index)
                            & (raw[cycle_index_hdr] == i),
                            "new_d",
                        ] = discharge_cap_last_val

                    good_cycles.append(i)

                except Exception:
                    bad_cycles.append(i)

            raw[charge_cap_hdr] = raw["new_c"]
            raw[discharge_cap_hdr] = raw["new_d"]
            raw.drop(["new_c", "new_d"], axis=1)
            if bad_cycles:
                logging.critical(f"The data contains bad cycles: {bad_cycles}")
        else:
            raise NotImplementedError(f"{capacity_structure} is not yet supported")

        step_time_conversion = self.structure.get("time_conversion_step_time", None)
        test_time_conversion = self.structure.get("time_conversion_test_time", None)
        date_time_conversion = self.structure.get("time_conversion_date_time", None)

        if date_time_conversion:
            if date_time_conversion.lower() == "test_time":
                raw[datetime_hdr] = raw[test_time_hdr]
                self.headers["datetime_txt"] = datetime_hdr
            else:
                raise NotImplementedError(
                    f"date_time conversion method not implemented ({date_time_conversion})"
                )

        if test_time_conversion:
            if test_time_conversion.lower() == "date_time_to_sec":
                start_time = raw[test_time_hdr].iloc[0]
                raw[test_time_hdr] = (
                    raw[test_time_hdr] - start_time
                ).dt.total_seconds()  # Warning: replaces original column
                raw[test_time_hdr] = raw[test_time_hdr]
            else:
                raise NotImplementedError(
                    f"test_time conversion method not implemented ({test_time_conversion})"
                )

        if step_time_conversion:
            if step_time_conversion.lower() == "time_to_sec":
                raw[step_time_hdr] = pd.to_timedelta(
                    raw[step_time_hdr]
                ).dt.total_seconds()
            else:
                raise NotImplementedError(
                    f"step_time conversion method not implemented ({step_time_conversion})"
                )
        raw = self._select_cols(raw)
        raw = self._check_cycleno_stepno(raw)

        data.raw_data_files_length.append(raw.shape[0])
        data.summary = None
        data.raw = raw
        new_tests.append(data)
        return new_tests

    def _select_cols(self, raw):
        selected = [
            self.headers_normal[col_def]
            for col_def in self.headers
            if self.headers_normal[col_def] in raw.columns
        ]
        raw = raw[selected]
        return raw

    def _parse_xls_data(self, file_name):
        sheet_name = self.structure["table_name"]

        raw_frame = pd.read_excel(
            file_name, engine="xlrd", sheet_name=None
        )  # TODO: replace this with pd.ExcelReader
        matching = [s for s in raw_frame.keys() if s.startswith(sheet_name)]
        if matching:
            return raw_frame[matching[0]]

    def _parse_xlsx_data(self, file_name):
        sheet_name = self.structure["table_name"]
        raw_frame = pd.read_excel(
            file_name, engine="openpyxl", sheet_name=None
        )  # TODO: replace this with pd.ExcelReader
        matching = [s for s in raw_frame.keys() if s.startswith(sheet_name)]
        if matching:
            return raw_frame[matching[0]]

    def _parse_csv_data(self, file_name, sep, header_row):
        raw = pd.read_csv(file_name, sep=sep, header=header_row, skip_blank_lines=False)
        return raw

    def _rename_cols(self, raw):
        rename_col_dict = dict()

        for col_def in self.headers:
            new_name = self.headers_normal[col_def]
            old_name = self.headers[col_def]
            if old_name in raw.columns:
                rename_col_dict[old_name] = new_name

        raw = raw.rename(columns=rename_col_dict)
        return raw

    # TODO: @jepe - finalize the _check sub-modules

    def _check_cycleno_stepno(self, raw):
        return raw

    def _convert_to_cellpy_units(self, data):
        return data

    def _check_columns(self, data):
        return data

    def _check_dtypes(self, data):
        return data

    def _generate_fid(self, file_name, var_dict):
        fid = FileID()
        last_modified = var_dict.get(self.variables["fid_last_modification_time"], None)
        size = var_dict.get(self.variables["fid_size"], None)
        last_accessed = var_dict.get(self.variables["fid_last_accessed"], None)

        if any([last_modified, size, last_accessed]):
            fid.name = os.path.abspath(file_name)
            fid.full_name = file_name
            fid.location = os.path.dirname(file_name)

            fid.size = size
            fid.last_modified = last_modified
            fid.last_accessed = last_accessed
            fid.last_info_changed = last_accessed
        else:
            fid.populate(file_name)

        return fid

    def inspect(self, data):
        data = self._convert_to_cellpy_units(data)
        data = self._check_columns(data)
        data = self._check_dtypes(data)
        return data

    def load(self, file_name):
        """Load a raw data-file

        Args:
            file_name (path)

        Returns:
            loaded test
        """

        new_rundata = self.loader(file_name)
        new_rundata = self.inspect(new_rundata)
        return new_rundata


def load_paal():
    import pathlib
    from pprint import pprint

    elkem_tester = pathlib.Path(
        r"C:\scripts\tasks\sibanode\2021_08_11_round_robbin\elkem_format.yml"
    )
    prms.Instruments.custom_instrument_definitions_file = elkem_tester
    print("running this")
    loader = CustomLoader()
    # loader.pick_definition_file()
    datadir = r"C:\scripts\tasks\sibanode\2021_08_11_round_robbin\data\raw\elkem\elkem"
    datadir = pathlib.Path(datadir)
    my_file_name = datadir / "240014-1-1-1621.xls"
    # print(help(loader.get_raw_units))
    # print(help(loader.get_raw_limits))
    # print(f"Trying to load {my_file_name}")
    loader.load(my_file_name)


def load_example_file():
    import pathlib
    from pprint import pprint

    print("running this")
    loader = CustomLoader()
    # loader.pick_definition_file()
    datadir = "/Users/jepe/scripting/cellpy/test_data"
    datadir = pathlib.Path(datadir)
    my_file_name = datadir / "custom_data_001.csv"
    print(help(loader.get_raw_units))
    print(help(loader.get_raw_limits))
    print(f"Trying to load {my_file_name}")
    loader.load(my_file_name)


if __name__ == "__main__":
    load_paal()



"""arbin res-type data files"""
import logging
import os
import platform
import shutil
import sys
import tempfile
import time
import warnings

import numpy as np
import pandas as pd

from cellpy import prms
from cellpy.parameters.internal_settings import get_headers_normal
from cellpy.readers.core import (
    Cell,
    FileID,
    check64bit,
    humanize_bytes,
    xldate_as_datetime,
)
from cellpy.readers.instruments.base import Loader

DEBUG_MODE = prms.Reader.diagnostics
ALLOW_MULTI_TEST_FILE = False

# Select odbc module
ODBC = prms._odbc
SEARCH_FOR_ODBC_DRIVERS = prms._search_for_odbc_driver

use_subprocess = prms.Instruments.Arbin.use_subprocess
detect_subprocess_need = prms.Instruments.Arbin.detect_subprocess_need

# Finding out some stuff about the platform
is_posix = False
is_macos = False
if os.name == "posix":
    is_posix = True
current_platform = platform.system()
if current_platform == "Darwin":
    is_macos = True

if DEBUG_MODE:
    logging.debug("DEBUG_MODE")
    logging.debug(f"ODBC: {ODBC}")
    logging.debug(f"SEARCH_FOR_ODBC_DRIVERS: {SEARCH_FOR_ODBC_DRIVERS}")
    logging.debug(f"use_subprocess: {use_subprocess}")
    logging.debug(f"detect_subprocess_need: {detect_subprocess_need}")
    logging.debug(f"current_platform: {current_platform}")

if detect_subprocess_need:
    logging.debug("detect_subprocess_need is True: checking versions")
    python_version, os_version = platform.architecture()
    if python_version == "64bit" and prms.Instruments.Arbin.office_version == "32bit":
        logging.debug(
            "python 64bit and office 32bit -> " "setting use_subprocess to True"
        )
        use_subprocess = True

if use_subprocess and not is_posix:
    # The windows users most likely have a strange custom path to mdbtools etc.
    logging.debug(
        "using subprocess (most lilkely mdbtools) " "on non-posix (most likely windows)"
    )
    if not prms.Instruments.Arbin.sub_process_path:
        sub_process_path = str(prms._sub_process_path)
    else:
        sub_process_path = str(prms.Instruments.Arbin.sub_process_path)

if is_posix:
    sub_process_path = "mdb-export"

try:
    driver_dll = prms.Instruments.Arbin.odbc_driver
except AttributeError:
    driver_dll = None

use_ado = False

if ODBC == "ado":
    use_ado = True
    logging.debug("Trying to use adodbapi as ado loader")
    try:
        import adodbapi as dbloader  # http://adodbapi.sourceforge.net/
    except ImportError:
        use_ado = False

if not use_ado:
    if ODBC == "pyodbc":
        try:
            import pyodbc as dbloader
        except ImportError:
            warnings.warn("COULD NOT LOAD DBLOADER!", ImportWarning)
            dbloader = None

    elif ODBC == "pypyodbc":
        try:
            import pypyodbc as dbloader
        except ImportError:
            warnings.warn("COULD NOT LOAD DBLOADER!", ImportWarning)
            dbloader = None

if DEBUG_MODE:
    logging.debug(f"dbloader: {dbloader}")

# The columns to choose if minimum selection is selected
MINIMUM_SELECTION = [
    "Data_Point",
    "Test_Time",
    "Step_Time",
    "DateTime",
    "Step_Index",
    "Cycle_Index",
    "Current",
    "Voltage",
    "Charge_Capacity",
    "Discharge_Capacity",
    "Internal_Resistance",
]

# Names of the tables in the .res db that is used by cellpy
TABLE_NAMES = {
    "normal": "Channel_Normal_Table",
    "global": "Global_Table",
    "statistic": "Channel_Statistic_Table",
}


normal_headers_renaming_dict = {
    "aci_phase_angle_txt": "ACI_Phase_Angle",
    "ref_aci_phase_angle_txt": "Reference_ACI_Phase_Angle",
    "ac_impedance_txt": "AC_Impedance",
    "ref_ac_impedance_txt": "Reference_AC_Impedance",
    "charge_capacity_txt": "Charge_Capacity",
    "charge_energy_txt": "Charge_Energy",
    "current_txt": "Current",
    "cycle_index_txt": "Cycle_Index",
    "data_point_txt": "Data_Point",
    "datetime_txt": "DateTime",
    "discharge_capacity_txt": "Discharge_Capacity",
    "discharge_energy_txt": "Discharge_Energy",
    "internal_resistance_txt": "Internal_Resistance",
    "is_fc_data_txt": "Is_FC_Data",
    "step_index_txt": "Step_Index",
    "sub_step_index_txt": "Sub_Step_Index",  # new
    "step_time_txt": "Step_Time",
    "sub_step_time_txt": "Sub_Step_Time",  # new
    "test_id_txt": "Test_ID",
    "test_time_txt": "Test_Time",
    "voltage_txt": "Voltage",
    "ref_voltage_txt": "Reference_Voltage",  # new
    "dv_dt_txt": "dV/dt",
    "frequency_txt": "Frequency",  # new
    "amplitude_txt": "Amplitude",  # new
}


class ArbinLoader(Loader):
    """Class for loading arbin-data from res-files."""

    def __init__(self, *args, **kwargs):
        """initiates the ArbinLoader class"""
        # could use __init__(self, cellpydata_object) and
        # set self.logger = cellpydata_object.logger etc.
        # then remember to include that as prm in "out of class" functions
        # self.prms = prms
        self.logger = logging.getLogger(__name__)
        # use the following prm to limit to loading only
        # one cycle or from cycle>x to cycle<x+n
        # prms.Reader.limit_loaded_cycles = [cycle from, cycle to]

        self.headers_normal = get_headers_normal()
        self.headers_global = self.get_headers_global()
        self.current_chunk = 0  # use this to set chunks to load

    @staticmethod
    def get_raw_units():
        raw_units = dict()
        raw_units["current"] = 1.0  # A
        raw_units["charge"] = 1.0  # Ah
        raw_units["mass"] = 0.001  # g
        return raw_units

    @staticmethod
    def get_headers_global():
        """Defines the so-called global column headings for Arbin .res-files"""
        headers = dict()
        # - global column headings (specific for Arbin)
        headers["applications_path_txt"] = "Applications_Path"
        headers["channel_index_txt"] = "Channel_Index"
        headers["channel_number_txt"] = "Channel_Number"
        headers["channel_type_txt"] = "Channel_Type"
        headers["comments_txt"] = "Comments"
        headers["creator_txt"] = "Creator"
        headers["daq_index_txt"] = "DAQ_Index"
        headers["item_id_txt"] = "Item_ID"
        headers["log_aux_data_flag_txt"] = "Log_Aux_Data_Flag"
        headers["log_chanstat_data_flag_txt"] = "Log_ChanStat_Data_Flag"
        headers["log_event_data_flag_txt"] = "Log_Event_Data_Flag"
        headers["log_smart_battery_data_flag_txt"] = "Log_Smart_Battery_Data_Flag"
        headers["mapped_aux_conc_cnumber_txt"] = "Mapped_Aux_Conc_CNumber"
        headers["mapped_aux_di_cnumber_txt"] = "Mapped_Aux_DI_CNumber"
        headers["mapped_aux_do_cnumber_txt"] = "Mapped_Aux_DO_CNumber"
        headers["mapped_aux_flow_rate_cnumber_txt"] = "Mapped_Aux_Flow_Rate_CNumber"
        headers["mapped_aux_ph_number_txt"] = "Mapped_Aux_PH_Number"
        headers["mapped_aux_pressure_number_txt"] = "Mapped_Aux_Pressure_Number"
        headers["mapped_aux_temperature_number_txt"] = "Mapped_Aux_Temperature_Number"
        headers["mapped_aux_voltage_number_txt"] = "Mapped_Aux_Voltage_Number"
        headers[
            "schedule_file_name_txt"
        ] = "Schedule_File_Name"  # KEEP FOR CELLPY FILE FORMAT
        headers["start_datetime_txt"] = "Start_DateTime"
        headers["test_id_txt"] = "Test_ID"  # KEEP FOR CELLPY FILE FORMAT
        headers["test_name_txt"] = "Test_Name"  # KEEP FOR CELLPY FILE FORMAT
        return headers

    @staticmethod
    def get_raw_limits():
        raw_limits = dict()
        raw_limits["current_hard"] = 0.000_000_000_000_1
        raw_limits["current_soft"] = 0.000_01
        raw_limits["stable_current_hard"] = 2.0
        raw_limits["stable_current_soft"] = 4.0
        raw_limits["stable_voltage_hard"] = 2.0
        raw_limits["stable_voltage_soft"] = 4.0
        raw_limits["stable_charge_hard"] = 0.001
        raw_limits["stable_charge_soft"] = 5.0
        raw_limits["ir_change"] = 0.00001
        return raw_limits

    def _get_res_connector(self, temp_filename):

        if use_ado:
            is64bit_python = check64bit(current_system="python")
            if is64bit_python:
                constr = (
                    "Provider=Microsoft.ACE.OLEDB.12.0; Data Source=%s" % temp_filename
                )
            else:
                constr = (
                    "Provider=Microsoft.Jet.OLEDB.4.0; Data Source=%s" % temp_filename
                )
            return constr

        if SEARCH_FOR_ODBC_DRIVERS:
            logging.debug("Searching for odbc drivers")
            try:
                drivers = [
                    driver
                    for driver in dbloader.drivers()
                    if "Microsoft Access Driver" in driver
                ]
                logging.debug(f"Found these: {drivers}")
                driver = drivers[0]

            except IndexError as e:
                logging.debug(
                    "Unfortunately, it seems the " "list of drivers is emtpy."
                )
                logging.debug("Use driver-name from config (if existing).")
                driver = driver_dll
                if is_macos:
                    driver = "/usr/local/lib/libmdbodbc.dylib"
                else:
                    if not driver:
                        print(
                            "\nCould not find any odbc-drivers suitable "
                            "for .res-type files. "
                            "Check out the homepage of pydobc for info on "
                            "installing drivers"
                        )
                        print(
                            "One solution that might work is downloading "
                            "the Microsoft Access database engine (in correct"
                            " bytes (32 or 64)) "
                            "from:\n"
                            "https://www.microsoft.com/en-us/download/"
                            "details.aspx?id=13255"
                        )
                        print(
                            "Or install mdbtools and set it up "
                            "(check the cellpy docs for help)"
                        )
                        print("\n")
                    else:
                        logging.debug("Using driver dll from config file")
                        logging.debug(f"driver dll: {driver}")

            self.logger.debug(f"odbc constr: {driver}")

        else:
            is64bit_python = check64bit(current_system="python")
            if is64bit_python:
                driver = "{Microsoft Access Driver (*.mdb, *.accdb)}"
            else:
                driver = "Microsoft Access Driver (*.mdb)"
            self.logger.debug("odbc constr: {}".format(driver))
        constr = "Driver=%s;Dbq=%s" % (driver, temp_filename)

        logging.debug(f"constr: {constr}")

        return constr

    def _clean_up_loadres(self, cur, conn, filename):
        if cur is not None:
            cur.close()  # adodbapi
        if conn is not None:
            conn.close()  # adodbapi
        if os.path.isfile(filename):
            try:
                os.remove(filename)
            except WindowsError as e:
                self.logger.warning("could not remove tmp-file\n%s %s" % (filename, e))

    def _post_process(self, data):
        fix_datetime = True
        set_index = True
        rename_headers = True  # could safely set this to false for now, but...

        # TODO:  insert post-processing and div tests here
        #    - check dtypes

        # Remark that we also set index during saving the file to hdf5 if
        #   it is not set.

        if rename_headers:
            columns = {}
            for key in self.headers_normal:
                old_header = normal_headers_renaming_dict[key]
                new_header = self.headers_normal[key]
                columns[old_header] = new_header

            data.raw.rename(index=str, columns=columns)

        if fix_datetime:
            h_datetime = self.headers_normal.datetime_txt
            logging.debug("converting to datetime format")
            data.raw[h_datetime] = data.raw[h_datetime].apply(
                xldate_as_datetime, option="to_datetime"
            )

            h_datetime = h_datetime
            if h_datetime in data.summary:
                data.summary[h_datetime] = data.summary[h_datetime].apply(
                    xldate_as_datetime, option="to_datetime"
                )

        if set_index:
            hdr_data_point = self.headers_normal.data_point_txt
            if data.raw.index.name != hdr_data_point:
                data.raw = data.raw.set_index(hdr_data_point, drop=False)

        return data

    def _inspect(self, run_data):
        """Inspect the file -> reports to log (debug)"""

        if not any([DEBUG_MODE]):
            return run_data

        if DEBUG_MODE:
            checked_rundata = []
            for data in run_data:
                new_cols = data.raw.columns
                for col in self.headers_normal:
                    if col not in new_cols:
                        logging.debug(f"Missing col: {col}")
                        # data.raw[col] = np.nan
                checked_rundata.append(data)
            return checked_rundata

    def _iterdump(self, file_name, headers=None):
        """
        Function for dumping values from a file.

        Should only be used by developers.

        Args:
            file_name: name of the file
            headers: list of headers to pick
                default:
                ["Discharge_Capacity", "Charge_Capacity"]

        Returns: pandas.DataFrame

        """
        if headers is None:
            headers = ["Discharge_Capacity", "Charge_Capacity"]

        step_txt = self.headers_normal.step_index_txt
        point_txt = self.headers_normal.data_point_txt
        cycle_txt = self.headers_normal.cycle_index_txt

        self.logger.debug("iterating through file: %s" % file_name)
        if not os.path.isfile(file_name):
            print("Missing file_\n   %s" % file_name)

        filesize = os.path.getsize(file_name)
        hfilesize = humanize_bytes(filesize)
        txt = "Filesize: %i (%s)" % (filesize, hfilesize)
        self.logger.info(txt)

        table_name_global = TABLE_NAMES["global"]
        table_name_stats = TABLE_NAMES["statistic"]
        table_name_normal = TABLE_NAMES["normal"]

        # creating temporary file and connection

        temp_dir = tempfile.gettempdir()
        temp_filename = os.path.join(temp_dir, os.path.basename(file_name))
        shutil.copy2(file_name, temp_dir)
        constr = self._get_res_connector(temp_filename)

        if use_ado:
            conn = dbloader.connect(constr)
        else:
            conn = dbloader.connect(constr, autocommit=True)

        self.logger.debug("tmp file: %s" % temp_filename)
        self.logger.debug("constr str: %s" % constr)

        # --------- read global-data ------------------------------------
        self.logger.debug("reading global data table")
        sql = "select * from %s" % table_name_global
        global_data_df = pd.read_sql_query(sql, conn)
        # col_names = list(global_data_df.columns.values)
        self.logger.debug("sql statement: %s" % sql)

        tests = global_data_df[self.headers_normal.test_id_txt]
        number_of_sets = len(tests)
        self.logger.debug("number of datasets: %i" % number_of_sets)
        self.logger.debug("only selecting first test")
        test_no = 0
        self.logger.debug("setting data for test number %i" % test_no)
        loaded_from = file_name
        # fid = FileID(file_name)
        start_datetime = global_data_df[self.headers_global["start_datetime_txt"]][
            test_no
        ]
        test_ID = int(global_data_df[self.headers_normal.test_id_txt][test_no])  # OBS
        test_name = global_data_df[self.headers_global["test_name_txt"]][test_no]

        # --------- read raw-data (normal-data) -------------------------
        self.logger.debug("reading raw-data")

        columns = ["Data_Point", "Step_Index", "Cycle_Index"]
        columns.extend(headers)
        columns_txt = ", ".join(["%s"] * len(columns)) % tuple(columns)

        sql_1 = "select %s " % columns_txt
        sql_2 = "from %s " % table_name_normal
        sql_3 = "where %s=%s " % (self.headers_normal.test_id_txt, test_ID)
        sql_5 = "order by %s" % self.headers_normal.data_point_txt
        import time

        info_list = []
        info_header = ["cycle", "row_count", "start_point", "end_point"]
        info_header.extend(headers)
        self.logger.info(" ".join(info_header))
        self.logger.info("-------------------------------------------------")

        for cycle_number in range(1, 2000):
            t1 = time.time()
            self.logger.debug("picking cycle %i" % cycle_number)
            sql_4 = "AND %s=%i " % (cycle_txt, cycle_number)
            sql = sql_1 + sql_2 + sql_3 + sql_4 + sql_5
            self.logger.debug("sql statement: %s" % sql)
            normal_df = pd.read_sql_query(sql, conn)
            t2 = time.time()
            dt = t2 - t1
            self.logger.debug("time: %f" % dt)
            if normal_df.empty:
                self.logger.debug("reached the end")
                break
            row_count, _ = normal_df.shape
            start_point = normal_df[point_txt].min()
            end_point = normal_df[point_txt].max()
            last = normal_df.iloc[-1, :]

            step_list = [cycle_number, row_count, start_point, end_point]
            step_list.extend([last[x] for x in headers])
            info_list.append(step_list)

        self._clean_up_loadres(None, conn, temp_filename)
        info_dict = pd.DataFrame(info_list, columns=info_header)
        return info_dict

    def investigate(self, file_name):
        """Investigate a .res file.

        Args:
            file_name: name of the file

        Returns: dictionary with div. stats and info.

        """
        step_txt = self.headers_normal.step_index_txt
        point_txt = self.headers_normal.data_point_txt
        cycle_txt = self.headers_normal.cycle_index_txt

        self.logger.debug("investigating file: %s" % file_name)
        if not os.path.isfile(file_name):
            print("Missing file_\n   %s" % file_name)

        filesize = os.path.getsize(file_name)
        hfilesize = humanize_bytes(filesize)
        txt = "Filesize: %i (%s)" % (filesize, hfilesize)
        self.logger.info(txt)

        table_name_global = TABLE_NAMES["global"]
        table_name_stats = TABLE_NAMES["statistic"]
        table_name_normal = TABLE_NAMES["normal"]

        # creating temporary file and connection

        temp_dir = tempfile.gettempdir()
        temp_filename = os.path.join(temp_dir, os.path.basename(file_name))
        shutil.copy2(file_name, temp_dir)
        constr = self._get_res_connector(temp_filename)

        if use_ado:
            conn = dbloader.connect(constr)
        else:
            conn = dbloader.connect(constr, autocommit=True)

        self.logger.debug("tmp file: %s" % temp_filename)
        self.logger.debug("constr str: %s" % constr)

        # --------- read global-data ------------------------------------
        self.logger.debug("reading global data table")
        sql = "select * from %s" % table_name_global
        global_data_df = pd.read_sql_query(sql, conn)
        # col_names = list(global_data_df.columns.values)
        self.logger.debug("sql statement: %s" % sql)

        tests = global_data_df[self.headers_normal.test_id_txt]
        number_of_sets = len(tests)
        self.logger.debug("number of datasets: %i" % number_of_sets)
        self.logger.debug("only selecting first test")
        test_no = 0
        self.logger.debug("setting data for test number %i" % test_no)
        loaded_from = file_name
        # fid = FileID(file_name)
        start_datetime = global_data_df[self.headers_global["start_datetime_txt"]][
            test_no
        ]
        test_ID = int(global_data_df[self.headers_normal.test_id_txt][test_no])  # OBS
        test_name = global_data_df[self.headers_global["test_name_txt"]][test_no]

        # --------- read raw-data (normal-data) -------------------------
        self.logger.debug("reading raw-data")

        columns = ["Data_Point", "Step_Index", "Cycle_Index"]
        columns_txt = ", ".join(["%s"] * len(columns)) % tuple(columns)

        sql_1 = "select %s " % columns_txt
        sql_2 = "from %s " % table_name_normal
        sql_3 = "where %s=%s " % (self.headers_normal.test_id_txt, test_ID)
        sql_5 = "order by %s" % self.headers_normal.data_point_txt
        import time

        info_list = []
        info_header = ["cycle", "step", "row_count", "start_point", "end_point"]
        self.logger.info(" ".join(info_header))
        self.logger.info("-------------------------------------------------")
        for cycle_number in range(1, 2000):
            t1 = time.time()
            self.logger.debug("picking cycle %i" % cycle_number)
            sql_4 = "AND %s=%i " % (cycle_txt, cycle_number)
            sql = sql_1 + sql_2 + sql_3 + sql_4 + sql_5
            self.logger.debug("sql statement: %s" % sql)
            normal_df = pd.read_sql_query(sql, conn)
            t2 = time.time()
            dt = t2 - t1
            self.logger.debug("time: %f" % dt)
            if normal_df.empty:
                self.logger.debug("reached the end")
                break
            row_count, _ = normal_df.shape
            steps = normal_df[self.headers_normal.step_index_txt].unique()
            txt = "cycle %i: %i [" % (cycle_number, row_count)
            for step in steps:
                self.logger.debug(" step: %i" % step)
                step_df = normal_df.loc[normal_df[step_txt] == step]
                step_row_count, _ = step_df.shape
                start_point = step_df[point_txt].min()
                end_point = step_df[point_txt].max()
                txt += " %i-(%i)" % (step, step_row_count)
                step_list = [cycle_number, step, step_row_count, start_point, end_point]
                info_list.append(step_list)

            txt += "]"
            self.logger.info(txt)

        self._clean_up_loadres(None, conn, temp_filename)
        info_dict = pd.DataFrame(info_list, columns=info_header)
        return info_dict

    def repair(self, file_name):
        """try to repair a broken/corrupted file"""
        raise NotImplemented

    def dump(self, file_name, path):
        """Dumps the raw file to an intermediate hdf5 file.

        This method can be used if the raw file is too difficult to load and it
        is likely that it is more efficient to convert it to an hdf5 format
        and then load it using the `from_intermediate_file` function.

        Args:
            file_name: name of the raw file
            path: path to where to store the intermediate hdf5 file (optional)

        Returns:
            full path to stored intermediate hdf5 file
            information about the raw file (needed by the
            `from_intermediate_file` function)

        """

        # information = None # contains information needed by the from_
        # intermediate_file reader
        # full_path = None
        # return full_path, information
        raise NotImplemented

    # def loader_win(
    #     self,
    #     temp_filename,
    #     table_name_global,
    #     file_name,
    #     bad_steps=None,
    #     dataset_number=None,
    #     data_points=None,
    # ):
    #
    #     constr = self._get_res_connector(temp_filename)
    #
    #     if use_ado:
    #         conn = dbloader.connect(constr)
    #     else:
    #         conn = dbloader.connect(constr, autocommit=True)
    #     self.logger.debug("constr str: %s" % constr)
    #
    #     self.logger.debug("reading global data table")
    #     sql = "select * from %s" % table_name_global
    #     self.logger.debug("sql statement: %s" % sql)
    #     global_data_df = pd.read_sql_query(sql, conn)
    #     # col_names = list(global_data_df.columns.values)
    #     tests = global_data_df[self.headers_normal.test_id_txt]
    #
    #     number_of_sets = len(tests)
    #     self.logger.debug("number of datasets: %i" % number_of_sets)
    #     self.logger.debug(f"datasets: {tests}")
    #
    #     if dataset_number is not None:
    #         self.logger.info(f"Dataset number given: {dataset_number}")
    #         self.logger.info(f"Available dataset numbers: {tests}")
    #         test_nos = [dataset_number]
    #     else:
    #         test_nos = range(number_of_sets)
    #
    #     for counter, test_no in enumerate(test_nos):
    #         if counter > 0:
    #             self.logger.warning("** WARNING ** MULTI-TEST-FILE (not recommended)")
    #             if not ALLOW_MULTI_TEST_FILE:
    #                 break
    #         data = self._init_data(file_name, global_data_df, test_no)
    #
    #         self.logger.debug("reading raw-data")
    #         if not use_mdbtools:
    #             # --------- read raw-data (normal-data) ------------------------
    #             length_of_test, normal_df = self._load_res_normal_table(
    #                 conn, data.test_ID, bad_steps
    #             )
    #             # --------- read stats-data (summary-data) ---------------------
    #             sql = "select * from %s where %s=%s order by %s" % (
    #                 table_name_stats,
    #                 self.headers_normal.test_id_txt,
    #                 data.test_ID,
    #                 self.headers_normal.data_point_txt,
    #             )
    #             summary_df = pd.read_sql_query(sql, conn)
    #
    #         else:
    #             length_of_test, normal_df, summary_df = self._load_from_tmp_files(
    #                 data, tmp_name_global, tmp_name_raw, tmp_name_stats, temp_filename
    #             )
    #
    #         if summary_df.empty and prms.Reader.use_cellpy_stat_file:
    #             txt = "\nCould not find any summary (stats-file)!"
    #             txt += "\n -> issue make_summary(use_cellpy_stat_file=False)"
    #             logging.debug(txt)
    #         # normal_df = normal_df.set_index("Data_Point")
    #
    #         data.summary = summary_df
    #         if DEBUG_MODE:
    #             mem_usage = normal_df.memory_usage()
    #             logging.debug(
    #                 f"memory usage for "
    #                 f"loaded data: \n{mem_usage}"
    #                 f"\ntotal: {humanize_bytes(mem_usage.sum())}"
    #             )
    #             logging.debug(f"time used: {(time.time() - time_0):2.4f} s")
    #
    #         data.raw = normal_df
    #         data.raw_data_files_length.append(length_of_test)
    #
    #         data = self._post_process(data)
    #         data = self.identify_last_data_point(data)
    #
    #         new_tests.append(data)
    #
    #     pass

    # def loader_posix(
    #     self, file_name, bad_steps=None, dataset_number=None, data_points=None
    # ):
    #     if is_posix:
    #         if is_macos:
    #             self.logger.debug("\nMAC OSX USING MDBTOOLS")
    #         else:
    #             self.logger.debug("\nPOSIX USING MDBTOOLS")
    #     else:
    #         self.logger.debug("\nWINDOWS USING MDBTOOLS-WIN")
    #
    #     tmp_name_global, tmp_name_raw, tmp_name_stats = self._create_tmp_files(
    #         table_name_global,
    #         table_name_normal,
    #         table_name_stats,
    #         temp_dir,
    #         temp_filename,
    #     )
    #
    #     # use pandas to load in the data
    #     global_data_df = pd.read_csv(tmp_name_global)
    #     tests = global_data_df[self.headers_normal.test_id_txt]
    #
    #     number_of_sets = len(tests)
    #     self.logger.debug("number of datasets: %i" % number_of_sets)
    #     self.logger.debug(f"datasets: {tests}")
    #
    #     if dataset_number is not None:
    #         self.logger.info(f"Dataset number given: {dataset_number}")
    #         self.logger.info(f"Available dataset numbers: {tests}")
    #         test_nos = [dataset_number]
    #     else:
    #         test_nos = range(number_of_sets)
    #
    #     for counter, test_no in enumerate(test_nos):
    #         if counter > 0:
    #             self.logger.warning("** WARNING ** MULTI-TEST-FILE (not recommended)")
    #             if not ALLOW_MULTI_TEST_FILE:
    #                 break
    #         data = self._init_data(file_name, global_data_df, test_no)
    #
    #         self.logger.debug("reading raw-data")
    #         if not use_mdbtools:
    #             # --------- read raw-data (normal-data) ------------------------
    #             length_of_test, normal_df = self._load_res_normal_table(
    #                 conn, data.test_ID, bad_steps
    #             )
    #             # --------- read stats-data (summary-data) ---------------------
    #             sql = "select * from %s where %s=%s order by %s" % (
    #                 table_name_stats,
    #                 self.headers_normal.test_id_txt,
    #                 data.test_ID,
    #                 self.headers_normal.data_point_txt,
    #             )
    #             summary_df = pd.read_sql_query(sql, conn)
    #
    #         else:
    #             length_of_test, normal_df, summary_df = self._load_from_tmp_files(
    #                 data, tmp_name_global, tmp_name_raw, tmp_name_stats, temp_filename
    #             )
    #
    #         if summary_df.empty and prms.Reader.use_cellpy_stat_file:
    #             txt = "\nCould not find any summary (stats-file)!"
    #             txt += "\n -> issue make_summary(use_cellpy_stat_file=False)"
    #             logging.debug(txt)
    #         # normal_df = normal_df.set_index("Data_Point")
    #
    #         data.summary = summary_df
    #         if DEBUG_MODE:
    #             mem_usage = normal_df.memory_usage()
    #             logging.debug(
    #                 f"memory usage for "
    #                 f"loaded data: \n{mem_usage}"
    #                 f"\ntotal: {humanize_bytes(mem_usage.sum())}"
    #             )
    #             logging.debug(f"time used: {(time.time() - time_0):2.4f} s")
    #
    #         data.raw = normal_df
    #         data.raw_data_files_length.append(length_of_test)
    #
    #         data = self._post_process(data)
    #         data = self.identify_last_data_point(data)
    #
    #         new_tests.append(data)
    #
    #     pass

    def loader(self, file_name, bad_steps=None, dataset_number=None, data_points=None):
        """Loads data from arbin .res files.

        Args:
            file_name (str): path to .res file.
            bad_steps (list of tuples): (c, s) tuples of steps s (in cycle c)
                to skip loading.
            dataset_number (int): the data set number to select if you are dealing
                with arbin files with more than one data-set.
            data_points (tuple of ints): load only data from data_point[0] to
                    data_point[1] (use None for infinite).

        Returns:
            new_tests (list of data objects)
        """
        # TODO: @jepe - insert kwargs - current chunk, only normal data, etc

        if DEBUG_MODE:
            time_0 = time.time()

        new_tests = []
        conn = None

        if not os.path.isfile(file_name):
            self.logger.info("Missing file_\n   %s" % file_name)
            return None

        self.logger.debug("in loader")
        self.logger.debug("filename: %s" % file_name)

        filesize = os.path.getsize(file_name)
        hfilesize = humanize_bytes(filesize)
        txt = "Filesize: %i (%s)" % (filesize, hfilesize)
        self.logger.debug(txt)
        if (
            filesize > prms.Instruments.Arbin.max_res_filesize
            and not prms.Reader.load_only_summary
        ):
            error_message = "\nERROR (loader):\n"
            error_message += "%s > %s - File is too big!\n" % (
                hfilesize,
                humanize_bytes(prms.Instruments.Arbin.max_res_filesize),
            )
            error_message += "(edit prms.Instruments.Arbin ['max_res_filesize'])\n"
            print(error_message)
            return None

        table_name_global = TABLE_NAMES["global"]
        table_name_stats = TABLE_NAMES["statistic"]
        table_name_normal = TABLE_NAMES["normal"]

        # creating temporary file and connection
        tmp_name_global = None
        tmp_name_raw = None
        tmp_name_stats = None

        temp_dir = tempfile.gettempdir()
        temp_filename = os.path.join(temp_dir, os.path.basename(file_name))
        shutil.copy2(file_name, temp_dir)
        self.logger.debug("tmp file: %s" % temp_filename)

        use_mdbtools = False
        if use_subprocess:
            use_mdbtools = True
        if is_posix:
            use_mdbtools = True

        # windows with same python bit as windows bit (the ideal case)

        # SPLIT FROM HERE
        if not use_mdbtools:
            constr = self._get_res_connector(temp_filename)

            if use_ado:
                conn = dbloader.connect(constr)
            else:
                conn = dbloader.connect(constr, autocommit=True)
            self.logger.debug("constr str: %s" % constr)

            self.logger.debug("reading global data table")
            sql = "select * from %s" % table_name_global
            self.logger.debug("sql statement: %s" % sql)
            global_data_df = pd.read_sql_query(sql, conn)
            # col_names = list(global_data_df.columns.values)

        else:
            if is_posix:
                if is_macos:
                    self.logger.debug("\nMAC OSX USING MDBTOOLS")
                else:
                    self.logger.debug("\nPOSIX USING MDBTOOLS")
            else:
                self.logger.debug("\nWINDOWS USING MDBTOOLS-WIN")

            tmp_name_global, tmp_name_raw, tmp_name_stats = self._create_tmp_files(
                table_name_global,
                table_name_normal,
                table_name_stats,
                temp_dir,
                temp_filename,
            )

            # use pandas to load in the data
            global_data_df = pd.read_csv(tmp_name_global)

        tests = global_data_df[self.headers_normal.test_id_txt]

        number_of_sets = len(tests)
        self.logger.debug("number of datasets: %i" % number_of_sets)
        self.logger.debug(f"datasets: {tests}")

        if dataset_number is not None:
            self.logger.info(f"Dataset number given: {dataset_number}")
            self.logger.info(f"Available dataset numbers: {tests}")
            test_nos = [dataset_number]
        else:
            test_nos = range(number_of_sets)

        for counter, test_no in enumerate(test_nos):
            if counter > 0:
                self.logger.warning("** WARNING ** MULTI-TEST-FILE (not recommended)")
                if not ALLOW_MULTI_TEST_FILE:
                    break
            data = self._init_data(file_name, global_data_df, test_no)

            self.logger.debug("reading raw-data")
            if not use_mdbtools:
                # --------- read raw-data (normal-data) ------------------------
                length_of_test, normal_df = self._load_res_normal_table(
                    conn, data.test_ID, bad_steps
                )
                # --------- read stats-data (summary-data) ---------------------
                sql = "select * from %s where %s=%s order by %s" % (
                    table_name_stats,
                    self.headers_normal.test_id_txt,
                    data.test_ID,
                    self.headers_normal.data_point_txt,
                )
                summary_df = pd.read_sql_query(sql, conn)

            else:
                length_of_test, normal_df, summary_df = self._load_from_tmp_files(
                    data, tmp_name_global, tmp_name_raw, tmp_name_stats, temp_filename
                )

            if summary_df.empty and prms.Reader.use_cellpy_stat_file:
                txt = "\nCould not find any summary (stats-file)!"
                txt += "\n -> issue make_summary(use_cellpy_stat_file=False)"
                logging.debug(txt)
            # normal_df = normal_df.set_index("Data_Point")

            data.summary = summary_df
            if DEBUG_MODE:
                mem_usage = normal_df.memory_usage()
                logging.debug(
                    f"memory usage for "
                    f"loaded data: \n{mem_usage}"
                    f"\ntotal: {humanize_bytes(mem_usage.sum())}"
                )
                logging.debug(f"time used: {(time.time() - time_0):2.4f} s")

            data.raw = normal_df
            data.raw_data_files_length.append(length_of_test)

            data = self._post_process(data)
            data = self.identify_last_data_point(data)

            new_tests.append(data)

        # COMBINE FROM HERE
        new_tests = self._inspect(new_tests)

        return new_tests

    def _create_tmp_files(
        self,
        table_name_global,
        table_name_normal,
        table_name_stats,
        temp_dir,
        temp_filename,
    ):
        import subprocess

        # creating tmp-filenames
        temp_csv_filename_global = os.path.join(temp_dir, "global_tmp.csv")
        temp_csv_filename_normal = os.path.join(temp_dir, "normal_tmp.csv")
        temp_csv_filename_stats = os.path.join(temp_dir, "stats_tmp.csv")
        # making the cmds
        mdb_prms = [
            (table_name_global, temp_csv_filename_global),
            (table_name_normal, temp_csv_filename_normal),
            (table_name_stats, temp_csv_filename_stats),
        ]
        # executing cmds
        for table_name, tmp_file in mdb_prms:
            with open(tmp_file, "w") as f:
                subprocess.call([sub_process_path, temp_filename, table_name], stdout=f)
                self.logger.debug(f"ran mdb-export {str(f)} {table_name}")
        return (
            temp_csv_filename_global,
            temp_csv_filename_normal,
            temp_csv_filename_stats,
        )

    def _load_from_tmp_files(
        self,
        data,
        temp_csv_filename_global,
        temp_csv_filename_normal,
        temp_csv_filename_stats,
        temp_filename,
    ):
        normal_df = pd.read_csv(temp_csv_filename_normal)
        # filter on test ID
        normal_df = normal_df[
            normal_df[self.headers_normal.test_id_txt] == data.test_ID
        ]
        # sort on data point
        if prms._sort_if_subprocess:
            normal_df = normal_df.sort_values(self.headers_normal.data_point_txt)
        length_of_test = normal_df.shape[0]
        summary_df = pd.read_csv(temp_csv_filename_stats)
        # clean up
        for f in [
            temp_filename,
            temp_csv_filename_stats,
            temp_csv_filename_normal,
            temp_csv_filename_global,
        ]:
            if os.path.isfile(f):
                try:
                    os.remove(f)
                except WindowsError as e:
                    self.logger.warning(f"could not remove tmp-file\n{f} {e}")
        return length_of_test, normal_df, summary_df

    def _init_data(self, file_name, global_data_df, test_no):
        data = Cell()
        data.cell_no = test_no
        data.loaded_from = file_name
        fid = FileID(file_name)
        # name of the .res file it is loaded from:
        # data.parent_filename = os.path.basename(file_name)
        data.channel_index = int(
            global_data_df[self.headers_global["channel_index_txt"]][test_no]
        )
        data.channel_number = int(
            global_data_df[self.headers_global["channel_number_txt"]][test_no]
        )
        data.creator = global_data_df[self.headers_global["creator_txt"]][test_no]
        data.item_ID = global_data_df[self.headers_global["item_id_txt"]][test_no]
        data.schedule_file_name = global_data_df[
            self.headers_global["schedule_file_name_txt"]
        ][test_no]
        data.start_datetime = global_data_df[self.headers_global["start_datetime_txt"]][
            test_no
        ]
        data.test_ID = int(global_data_df[self.headers_normal.test_id_txt][test_no])
        data.test_name = global_data_df[self.headers_global["test_name_txt"]][test_no]
        data.raw_data_files.append(fid)
        return data

    def _normal_table_generator(self, **kwargs):
        pass

    def _load_res_normal_table(self, conn, test_ID, bad_steps):
        # Note that this function is run each time you use the loader.
        # This means that it is not ideal for
        # handling generators etc

        self.logger.debug("starting loading raw-data")
        self.logger.debug(f"connection: {conn} test-ID: {test_ID}")
        self.logger.debug(f"bad steps:  {bad_steps}")

        table_name_normal = TABLE_NAMES["normal"]

        if prms.Reader.load_only_summary:  # SETTING
            warnings.warn("not implemented")

        if prms.Reader.select_minimal:  # SETTING
            columns = MINIMUM_SELECTION
            columns_txt = ", ".join(["%s"] * len(columns)) % tuple(columns)
        else:
            columns_txt = "*"

        sql_1 = "select %s " % columns_txt
        sql_2 = "from %s " % table_name_normal
        sql_3 = "where %s=%s " % (self.headers_normal.test_id_txt, test_ID)
        sql_4 = ""

        if bad_steps is not None:
            if not isinstance(bad_steps, (list, tuple)):
                bad_steps = [bad_steps]
            for bad_cycle, bad_step in bad_steps:
                self.logger.debug(f"bad_step def: [c={bad_cycle}, s={bad_step}]")
                sql_4 += "AND NOT (%s=%i " % (
                    self.headers_normal.cycle_index_txt,
                    bad_cycle,
                )
                sql_4 += "AND %s=%i) " % (self.headers_normal.step_index_txt, bad_step)

        if prms.Reader.limit_loaded_cycles:
            if len(prms.Reader.limit_loaded_cycles) > 1:
                sql_4 += "AND %s>%i " % (
                    self.headers_normal.cycle_index_txt,
                    prms.Reader.limit_loaded_cycles[0],
                )
                sql_4 += "AND %s<%i " % (
                    self.headers_normal.cycle_index_txt,
                    prms.Reader.limit_loaded_cycles[-1],
                )
            else:
                sql_4 = "AND %s=%i " % (
                    self.headers_normal.cycle_index_txt,
                    prms.Reader.limit_loaded_cycles[0],
                )

        sql_5 = "order by %s" % self.headers_normal.data_point_txt
        sql = sql_1 + sql_2 + sql_3 + sql_4 + sql_5

        self.logger.debug("INFO ABOUT LOAD RES NORMAL")
        self.logger.debug("sql statement: %s" % sql)

        if DEBUG_MODE:
            current_memory_usage = sys.getsizeof(self)

        if not prms.Instruments.Arbin.chunk_size:
            self.logger.debug("no chunk-size given")
            # memory here
            normal_df = pd.read_sql_query(sql, conn)
            # memory here
            length_of_test = normal_df.shape[0]
            self.logger.debug(f"loaded to normal_df (length =  {length_of_test})")
        else:
            self.logger.debug(f"chunk-size: {prms.Instruments.Arbin.chunk_size}")
            self.logger.debug("creating a pd.read_sql_query generator")
            normal_df_reader = pd.read_sql_query(
                sql, conn, chunksize=prms.Instruments.Arbin.chunk_size
            )
            normal_df = None
            chunk_number = 0
            self.logger.debug("created pandas sql reader")
            self.logger.debug("iterating chunk-wise")
            for i, chunk in enumerate(normal_df_reader):
                self.logger.debug(f"iteration number {i}")
                if prms.Instruments.Arbin.max_chunks:
                    self.logger.debug(
                        f"max number of chunks mode "
                        f"({prms.Instruments.Arbin.max_chunks})"
                    )
                    if chunk_number < prms.Instruments.Arbin.max_chunks:
                        normal_df = pd.concat([normal_df, chunk], ignore_index=True)
                        self.logger.debug(
                            f"chunk {i} of {prms.Instruments.Arbin.max_chunks}"
                        )
                    else:
                        break
                else:
                    try:
                        normal_df = pd.concat([normal_df, chunk], ignore_index=True)
                        self.logger.debug("concatenated new chunk")
                    except MemoryError:
                        self.logger.error(
                            " - Could not read complete file (MemoryError)."
                        )
                        self.logger.error(
                            f"Last successfully loaded chunk " f"number: {chunk_number}"
                        )
                        self.logger.error(
                            f"Chunk size: {prms.Instruments.Arbin.chunk_size}"
                        )
                        break
                chunk_number += 1
            length_of_test = normal_df.shape[0]
            self.logger.debug(f"finished iterating (#rows: {length_of_test})")
        return length_of_test, normal_df


if __name__ == "__main__":
    import logging

    from cellpy import log

    log.setup_logging(default_level="DEBUG")



""" This module contains several of the most important classes used in cellpy.

It also contains functions that are used by readers and utils. And it has the file-
version definitions.
"""

import collections
import datetime
import importlib
import logging
import os
import pathlib
import pickle
import sys
import time
import warnings
from functools import wraps

import numpy as np
import pandas as pd
from scipy import interpolate

from cellpy.exceptions import NullData
from cellpy.parameters import prms
from cellpy.parameters.internal_settings import (
    ATTRS_CELLPYFILE,
    cellpy_limits,
    cellpy_units,
    get_headers_normal,
    get_headers_step_table,
    get_headers_summary,
)

CELLPY_FILE_VERSION = 6
MINIMUM_CELLPY_FILE_VERSION = 4
STEP_TABLE_VERSION = 5
RAW_TABLE_VERSION = 5
SUMMARY_TABLE_VERSION = 5
PICKLE_PROTOCOL = 4

HEADERS_NORMAL = get_headers_normal()
HEADERS_SUMMARY = get_headers_summary()
HEADERS_STEP_TABLE = get_headers_step_table()


# https://stackoverflow.com/questions/60067953/
# 'is-it-possible-to-specify-the-pickle-protocol-when-writing-pandas-to-hdf5
class PickleProtocol:
    def __init__(self, level):
        self.previous = pickle.HIGHEST_PROTOCOL
        self.level = level

    def __enter__(self):
        importlib.reload(pickle)
        pickle.HIGHEST_PROTOCOL = self.level

    def __exit__(self, *exc):
        importlib.reload(pickle)
        pickle.HIGHEST_PROTOCOL = self.previous


def pickle_protocol(level):
    return PickleProtocol(level)


class FileID:
    """class for storing information about the raw-data files.

    This class is used for storing and handling raw-data file information.
    It is important to keep track of when the data was extracted from the
    raw-data files so that it is easy to know if the hdf5-files used for
    @storing "treated" data is up-to-date.

    Attributes:
        name (str): Filename of the raw-data file.
        full_name (str): Filename including path of the raw-data file.
        size (float): Size of the raw-data file.
        last_modified (datetime): Last time of modification of the raw-data
            file.
        last_accessed (datetime): last time of access of the raw-data file.
        last_info_changed (datetime): st_ctime of the raw-data file.
        location (str): Location of the raw-data file.

    """

    def __init__(self, filename=None):
        make_defaults = True
        if filename:
            if os.path.isfile(filename):
                fid_st = os.stat(filename)
                self.name = os.path.abspath(filename)
                self.full_name = filename
                self.size = fid_st.st_size
                self.last_modified = fid_st.st_mtime
                self.last_accessed = fid_st.st_atime
                self.last_info_changed = fid_st.st_ctime
                self.location = os.path.dirname(filename)
                self.last_data_point = 0  # used later when updating is implemented
                make_defaults = False

        if make_defaults:
            self.name = None
            self.full_name = None
            self.size = 0
            self.last_modified = None
            self.last_accessed = None
            self.last_info_changed = None
            self.location = None
            self._last_data_point = 0  # to be used later when updating is implemented

    def __str__(self):
        txt = "\n<fileID>\n"
        txt += f"full name: {self.full_name}\n"
        txt += f"name: {self.name}\n"
        txt += f"location: {self.location}\n"
        if self.last_modified is not None:
            txt += f"modified: {self.last_modified}\n"
        else:
            txt += "modified: NAN\n"
        if self.size is not None:
            txt += f"size: {self.size}\n"
        else:
            txt += "size: NAN\n"

        txt += f"last data point: {self.last_data_point}\n"
        return txt

    @property
    def last_data_point(self):
        # TODO: consider including a method here to find the last data point (raw data)
        # ideally, this value should be set when loading the raw data before
        # merging files (if it consists of several files)
        return self._last_data_point

    @last_data_point.setter
    def last_data_point(self, value):
        self._last_data_point = value

    def populate(self, filename):
        """Finds the file-stats and populates the class with stat values.

        Args:
            filename (str): name of the file.
        """

        if os.path.isfile(filename):
            fid_st = os.stat(filename)
            self.name = os.path.abspath(filename)
            self.full_name = filename
            self.size = fid_st.st_size
            self.last_modified = fid_st.st_mtime
            self.last_accessed = fid_st.st_atime
            self.last_info_changed = fid_st.st_ctime
            self.location = os.path.dirname(filename)

    def get_raw(self):
        """Get a list with information about the file.

        The returned list contains name, size, last_modified and location.
        """
        return [self.name, self.size, self.last_modified, self.location]

    def get_name(self):
        """Get the filename."""
        return self.name

    def get_size(self):
        """Get the size of the file."""
        return self.size

    def get_last(self):
        """Get last modification time of the file."""
        return self.last_modified


class Cell:
    """Object to store data for a test.

    This class is used for storing all the relevant data for a 'run', i.e. all
    the data collected by the tester as stored in the raw-files.

    Attributes:
        test_no (int): test number.
        mass (float): mass of electrode [mg].
        dfdata (pandas.DataFrame): contains the experimental data points.
        dfsummary (pandas.DataFrame): contains summary of the data pr. cycle.
        step_table (pandas.DataFrame): information for each step, used for
            defining type of step (charge, discharge, etc.)

    """

    def _repr_html_(self):
        obj = f"<b>Cell-object</b> id={hex(id(self))}"
        txt = "<p>"
        for p in dir(self):
            if not p.startswith("_"):
                if p not in ["raw", "summary", "steps", "logger"]:
                    value = self.__getattribute__(p)
                    txt += f"<b>{p}</b>: {value}<br>"
        txt += "</p>"
        try:
            raw_txt = f"<p><b>raw data-frame (summary)</b><br>{self.raw.describe()._repr_html_()}</p>"
            raw_txt += f"<p><b>raw data-frame (head)</b><br>{self.raw.head()._repr_html_()}</p>"
        except AttributeError:
            raw_txt = "<p><b>raw data-frame </b><br> not found!</p>"
        except ValueError:
            raw_txt = "<p><b>raw data-frame </b><br> does not contain any columns!</p>"

        try:
            summary_txt = f"<p><b>summary data-frame (summary)</b><br>{self.summary.describe()._repr_html_()}</p>"
            summary_txt += f"<p><b>summary data-frame (head)</b><br>{self.summary.head()._repr_html_()}</p>"
        except AttributeError:
            summary_txt = "<p><b>summary data-frame </b><br> not found!</p>"
        except ValueError:
            summary_txt = (
                "<p><b>summary data-frame </b><br> does not contain any columns!</p>"
            )

        try:
            steps_txt = f"<p><b>steps data-frame (summary)</b><br>{self.steps.describe()._repr_html_()}</p>"
            steps_txt += f"<p><b>steps data-frame (head)</b><br>{self.steps.head()._repr_html_()}</p>"
        except AttributeError:
            steps_txt = "<p><b>steps data-frame </b><br> not found!</p>"
        except ValueError:
            steps_txt = (
                "<p><b>steps data-frame </b><br> does not contain any columns!</p>"
            )

        return obj + txt + summary_txt + steps_txt + raw_txt

    def __init__(self, **kwargs):
        self.logger = logging.getLogger(__name__)
        self.logger.debug("created DataSet instance")

        # meta-data
        self.cell_no = None
        self.mass = prms.Materials.default_mass  # active material (in mg)
        self.tot_mass = prms.Materials.default_mass  # total material (in mg)
        self.no_cycles = 0.0
        self.charge_steps = None
        self.discharge_steps = None
        self.ir_steps = None
        self.ocv_steps = None
        self.nom_cap = prms.DataSet.nom_cap  # mAh/g (for finding c-rates)
        self.mass_given = False
        self.material = prms.Materials.default_material
        self.merged = False
        self.file_errors = None  # not in use at the moment
        self.loaded_from = None  # loaded from (can be list if merged)
        self.channel_index = None
        self.channel_number = None
        self.creator = None
        self.item_ID = None
        self.schedule_file_name = None
        self.start_datetime = None
        self.test_ID = None
        self.name = None

        # new meta data
        self.cycle_mode = prms.Reader.cycle_mode
        self.active_electrode_area = None  # [cm2]
        self.active_electrode_thickness = None  # [micron]
        self.electrolyte_type = None  #
        self.electrolyte_volume = None  # [micro-liter]
        self.active_electrode_type = None
        self.counter_electrode_type = None
        self.reference_electrode_type = None
        self.experiment_type = None
        self.cell_type = None
        self.separator_type = None
        self.active_electrode_current_collector = None
        self.reference_electrode_current_collector = None
        self.comment = None

        # custom meta-data
        for k in kwargs:
            if hasattr(self, k):
                setattr(self, k, kwargs[k])

        # methods in CellpyData to update if adding new attributes:
        # ATTRS_CELLPYFILE

        # place to put "checks" etc:
        # _extract_meta_from_cellpy_file
        # _create_infotable()

        self.raw_data_files = []
        self.raw_data_files_length = []
        self.raw_units = cellpy_units
        self.raw_limits = cellpy_limits

        # self.data = collections.OrderedDict()  # not used
        # self.summary = collections.OrderedDict()  # not used

        self.raw = pd.DataFrame()
        self.summary = pd.DataFrame()
        # self.summary_made = False  # Should be removed
        self.steps = pd.DataFrame()  # is this used? - check!
        # self.step_table_made = False  # Should be removed
        # self.parameter_table = collections.OrderedDict()
        self.summary_table_version = SUMMARY_TABLE_VERSION
        self.step_table_version = STEP_TABLE_VERSION
        self.cellpy_file_version = CELLPY_FILE_VERSION
        self.raw_table_version = RAW_TABLE_VERSION

    @staticmethod
    def _header_str(hdr):
        txt = "\n"
        txt += 80 * "-" + "\n"
        txt += f" {hdr} ".center(80) + "\n"
        txt += 80 * "-" + "\n"
        return txt

    def __str__(self):
        txt = "<DataSet>\n"
        txt += "loaded from file\n"
        if isinstance(self.loaded_from, (list, tuple)):
            for f in self.loaded_from:
                txt += str(f)
                txt += "\n"

        else:
            txt += str(self.loaded_from)
            txt += "\n"
        txt += "\n* GLOBAL\n"
        txt += f"material:            {self.material}\n"
        txt += f"mass (active):       {self.mass}\n"
        txt += f"test ID:             {self.test_ID}\n"
        txt += f"mass (total):        {self.tot_mass}\n"
        txt += f"nominal capacity:    {self.nom_cap}\n"
        txt += f"channel index:       {self.channel_index}\n"
        txt += f"DataSet name:        {self.name}\n"
        txt += f"creator:             {self.creator}\n"
        txt += f"schedule file name:  {self.schedule_file_name}\n"

        try:
            if self.start_datetime:
                start_datetime_str = xldate_as_datetime(self.start_datetime)
            else:
                start_datetime_str = "Not given"
        except AttributeError:
            start_datetime_str = "NOT READABLE YET"

        txt += f"start-date:         {start_datetime_str}\n"

        txt += self._header_str("DATA")
        try:
            txt += str(self.raw.describe())
        except (AttributeError, ValueError):
            txt += "EMPTY (Not processed yet)\n"

        txt += self._header_str("SUMMARY")
        try:
            txt += str(self.summary.describe())
        except (AttributeError, ValueError):
            txt += "EMPTY (Not processed yet)\n"

        txt += self._header_str("STEP TABLE")
        try:
            txt += str(self.steps.describe())
            txt += str(self.steps.head())
        except (AttributeError, ValueError):
            txt += "EMPTY (Not processed yet)\n"

        txt += self._header_str("RAW UNITS")
        try:
            txt += str(self.raw.describe())
            txt += str(self.raw.head())
        except (AttributeError, ValueError):
            txt += "EMPTY (Not processed yet)\n"
        return txt

    @property
    def summary_made(self):
        """check if the summary table exists"""
        try:
            empty = self.summary.empty
        except AttributeError:
            empty = True
        return not empty

    @property
    def steps_made(self):
        """check if the step table exists"""
        try:
            empty = self.steps.empty
        except AttributeError:
            empty = True
        return not empty

    @property
    def no_data(self):
        # TODO: @jepe should consider renaming this to be in-line with "steps_made" etc. (or renaming steps_made and
        #   summary_made to e.g. no_steps, no_summary)
        try:
            empty = self.raw.empty
        except AttributeError:
            empty = True
        return empty


class ObjectFactory:
    def __init__(self):
        self._builders = {}
        self._kwargs = {}

    def register_builder(self, key, builder, **kwargs):
        logging.debug(f"Registering instrument {key}")
        self._builders[key] = builder
        self._kwargs = kwargs

    def create(self, key, **kwargs):
        builder = self._builders.get(key)
        if not builder:
            raise ValueError(key)
        return builder(**kwargs)


def register_instruments():
    # Example for future use
    # This should be moved to cellreader
    factory = ObjectFactory()
    instruments = find_all_instruments()
    for instrument_id, instrument in instruments.items():
        factory.register_builder(instrument_id, instrument)


def _extract_loader_class_from_module(module):
    # insert some magic here (find sub-class of BaseLoader)
    return module


def find_all_instruments():
    # Example for future use.
    # Keeping imports here so that it will simplify future
    #   refactoring.
    # In addition, this functions needs to implement a way to
    # get appropriate names of the different loaders. Or maybe
    # tweak the base loader etc so that it can get the appropriate
    # name from foo.__name__.

    # This function is not ready for use yet. One important part
    # still missing is to find the actual class inside the module
    # (the .py file) that should be loaded. It is probably not
    # very difficult to implement (since it is a subclass of BaseLoader).

    # Another missing part is how to find externally installed loaders
    # (plugins). Of obvious reasons; plugins are not supported yet.

    # Also, need to properly utilise and propagate the instrument loader
    # names etc (using query_instrument).

    from importlib.machinery import SourceFileLoader
    import cellpy.readers.instruments.configurations as site_1
    import cellpy.readers.instruments as site_2

    instruments = {}
    logging.debug("Searches for modules in configurations folder:")

    site_1 = pathlib.Path(site_1.__file__).parent
    modules_in_site_1 = [
        s for s in site_1.glob("*.py") if not str(s.name).startswith("_")
    ]

    for module in modules_in_site_1:
        module_name = module.name.rstrip(".py")
        foo = SourceFileLoader(module_name, str(module)).load_module()
        instrument_name = foo.__name__
        instrument_class = _extract_loader_class_from_module(foo)
        instruments[instrument_name] = instrument_class
        logging.debug(instrument_name)

    logging.debug("Searching for modules in base instrument folder:")

    site_2 = pathlib.Path(site_2.__file__).parent
    modules_in_site_2 = [
        s
        for s in site_2.glob("*.py")
        if not (
            str(s.name).startswith("_")
            or str(s.name).startswith("dev_")
            or str(s.name).startswith("base")
            or str(s.name).startswith("backup")
        )
    ]

    for module in modules_in_site_2:
        module_name = module.name.rstrip(".py")
        foo = SourceFileLoader(module_name, str(module)).load_module()
        instrument_name = foo.__name__
        instrument_class = _extract_loader_class_from_module(foo)
        instruments[instrument_name] = instrument_class
        logging.debug(instrument_name)

    logging.debug("Searching for module configurations " "in user instrument folder:")
    # These are only yaml-files and should ideally import the appropriate
    #    custom loader class
    logging.debug("- Not implemented yet")

    logging.debug("Searching for modules through plug-ins:")
    # Not sure how to do this yet. Probably also some importlib trick.
    logging.debug("- Not implemented yet")
    return instruments


def __look_up_instrument(instrument):
    if instrument in ["arbin", "arbin_res"]:
        from cellpy.readers.instruments.arbin_res import ArbinLoader as RawLoader

        instrument_id = "arbin"
    elif instrument == "arbin_sql":
        from cellpy.readers.instruments.arbin_sql import ArbinSQLLoader as RawLoader

        instrument_id = "arbin_sql"
    elif instrument == "arbin_sql_csv":
        from cellpy.readers.instruments.arbin_sql_csv import (
            ArbinCsvLoader as RawLoader,
        )

        instrument_id = "arbin_sql_csv"
    elif instrument == "arbin_sql_xlsx":
        from cellpy.readers.instruments.arbin_sql_xlsx import (
            ArbinXLSXLoader as RawLoader,
        )

        instrument_id = "arbin_sql_xlsx"

    elif instrument in ["pec", "pec_csv"]:
        from cellpy.readers.instruments.pec import PECLoader as RawLoader

        instrument_id = "pec"

    elif instrument in ["biologics", "biologics_mpr"]:
        from cellpy.readers.instruments.biologics_mpr import MprLoader as RawLoader

        instrument_id = "biologics"

    elif instrument in ["maccor", "maccor_txt"]:
        from cellpy.readers.instruments.maccor_txt import (
            MaccorTxtLoader as RawLoader,
        )

        instrument_id = "maccor"
        # need more here (model etc)

    elif instrument.startswith("custom"):
        from cellpy.readers.instruments.custom_instrument import (
            CustomTxtLoader as RawLoader,
        )

        instrument_id = "custom"

    elif instrument.startswith("old_custom"):
        print("OLD CUSTOM LOADER")
        from cellpy.readers.instruments.custom import CustomLoader as RawLoader

        instrument_id = "old_custom"

    elif instrument.endswith(".yml"):
        from cellpy.readers.instruments.local_instrument import (
            LocalTxtLoader as RawLoader,
        )

        instrument_id = instrument
        # fix this

    else:
        raise Exception(f"option does not exist: '{instrument}'")

    return RawLoader, instrument_id


def query_instrument(variable, instrument=None, instrument_file=None, **kwargs):
    """Retrieve information from a loader class without instantiating it.

    Remark! This function uses the .get_params method for the loader class and
        not all loaders have this method implemented. This function will catch
        several exceptions (`AttributeError`, `NotImplementedError`, `KeyError`)
        without propagating it. Thus, it is usually OK to use this function, but
        you might not get anything else than `None` from it.
    """

    RawLoader, instrument_id = __look_up_instrument(instrument)
    try:
        value = RawLoader.get_params(variable)
        logging.debug(f"GOT {variable}={value} for {instrument}")
        return value

    except (AttributeError, NotImplementedError, KeyError):
        logging.debug(f"COULD NOT RETRIEVE {variable} for {instrument}")


def identify_last_data_point(data):
    """Find the last data point and store it in the fid instance"""

    logging.debug("searching for last data point")
    hdr_data_point = HEADERS_NORMAL.data_point_txt
    try:
        if hdr_data_point in data.raw.columns:

            last_data_point = data.raw[hdr_data_point].max()
        else:
            last_data_point = data.raw.index.max()
    except AttributeError:
        logging.debug("AttributeError - setting last data point to 0")
        last_data_point = 0
    if not last_data_point > 0:
        last_data_point = 0
    data.raw_data_files[0].last_data_point = last_data_point
    logging.debug(f"last data point: {last_data_point}")
    return data


def check64bit(current_system="python"):
    """checks if you are on a 64 bit platform"""
    if current_system == "python":
        return sys.maxsize > 2147483647
    elif current_system == "os":
        import platform

        pm = platform.machine()
        if pm != ".." and pm.endswith("64"):  # recent Python (not Iron)
            return True
        else:
            if "PROCESSOR_ARCHITEW6432" in os.environ:
                return True  # 32 bit program running on 64 bit Windows
            try:
                # 64 bit Windows 64 bit program
                return os.environ["PROCESSOR_ARCHITECTURE"].endswith("64")
            except IndexError:
                pass  # not Windows
            try:
                # this often works in Linux
                return "64" in platform.architecture()[0]
            except Exception:
                # is an older version of Python, assume also an older os@
                # (best we can guess)
                return False


def humanize_bytes(b, precision=1):
    """Return a humanized string representation of a number of b."""

    abbrevs = (
        (1 << 50, "PB"),
        (1 << 40, "TB"),
        (1 << 30, "GB"),
        (1 << 20, "MB"),
        (1 << 10, "kB"),
        (1, "b"),
    )
    if b == 1:
        return "1 byte"
    for factor, suffix in abbrevs:
        if b >= factor:
            break
    # return '%.*f %s' % (precision, old_div(b, factor), suffix)
    return "%.*f %s" % (precision, b // factor, suffix)


def xldate_as_datetime(xldate, datemode=0, option="to_datetime"):
    """Converts a xls date stamp to a more sensible format.

    Args:
        xldate (str): date stamp in Excel format.
        datemode (int): 0 for 1900-based, 1 for 1904-based.
        option (str): option in ("to_datetime", "to_float", "to_string"),
            return value

    Returns:
        datetime (datetime object, float, or string).

    """

    # This does not work for numpy-arrays

    if option == "to_float":
        d = (xldate - 25589) * 86400.0
    else:
        try:
            d = datetime.datetime(1899, 12, 30) + datetime.timedelta(
                days=xldate + 1462 * datemode
            )
            # date_format = "%Y-%m-%d %H:%M:%S:%f" # with microseconds,
            # excel cannot cope with this!
            if option == "to_string":
                date_format = "%Y-%m-%d %H:%M:%S"  # without microseconds
                d = d.strftime(date_format)
        except TypeError:
            logging.info(f"The date is not of correct type [{xldate}]")
            d = xldate
    return d


def convert_to_mAhg(c, mass=1.0):
    """Converts capacity in Ah to capacity in mAh/g.

    Args:
        c (float or numpy array): capacity in mA.
        mass (float): mass in mg.

    Returns:
        float: 1000000 * c / mass
    """
    return 1_000_000 * c / mass


def collect_ocv_curves():
    raise NotImplementedError


def collect_capacity_curves(
    data,
    direction="charge",
    trim_taper_steps=None,
    steps_to_skip=None,
    steptable=None,
    max_cycle_number=None,
    **kwargs,
):
    """Create a list of pandas.DataFrames, one for each charge step.

    The DataFrames are named by its cycle number.

    Input: CellpyData
    Returns: list of pandas.DataFrames,
        list of cycle numbers,
        minimum voltage value,
        maximum voltage value"""

    # TODO: should allow for giving cycle numbers as input (e.g. cycle=[1, 2, 10]
    #  or cycle=2), not only max_cycle_number

    minimum_v_value = np.Inf
    maximum_v_value = -np.Inf
    charge_list = []
    cycles = kwargs.pop("cycle", None)

    if cycles is None:
        cycles = data.get_cycle_numbers()

    if max_cycle_number is None:
        max_cycle_number = max(cycles)

    for cycle in cycles:
        if cycle > max_cycle_number:
            break
        try:
            if direction == "charge":
                q, v = data.get_ccap(
                    cycle,
                    trim_taper_steps=trim_taper_steps,
                    steps_to_skip=steps_to_skip,
                    steptable=steptable,
                )
            else:
                q, v = data.get_dcap(
                    cycle,
                    trim_taper_steps=trim_taper_steps,
                    steps_to_skip=steps_to_skip,
                    steptable=steptable,
                )

        except NullData as e:
            logging.warning(e)
            d = pd.DataFrame()
            d.name = cycle
            charge_list.append(d)
        else:
            d = pd.DataFrame({"q": q, "v": v})
            # d.name = f"{cycle}"
            d.name = cycle
            charge_list.append(d)
            v_min = v.min()
            v_max = v.max()
            if v_min < minimum_v_value:
                minimum_v_value = v_min
            if v_max > maximum_v_value:
                maximum_v_value = v_max
    return charge_list, cycles, minimum_v_value, maximum_v_value


def interpolate_y_on_x(
    df,
    x=None,
    y=None,
    new_x=None,
    dx=10.0,
    number_of_points=None,
    direction=1,
    **kwargs,
):
    """Interpolate a column based on another column.

    Args:
        df: DataFrame with the (cycle) data.
        x: Column name for the x-value (defaults to the step-time column).
        y: Column name for the y-value (defaults to the voltage column).
        new_x (numpy array or None): Interpolate using these new x-values
            instead of generating x-values based on dx or number_of_points.
        dx: step-value (defaults to 10.0)
        number_of_points: number of points for interpolated values (use
            instead of dx and overrides dx if given).
        direction (-1,1): if direction is negative, then invert the
            x-values before interpolating.
        **kwargs: arguments passed to scipy.interpolate.interp1d

    Returns: DataFrame with interpolated y-values based on given or
        generated x-values.

    """

    # TODO: allow for giving a fixed interpolation range (x-values).
    #  Remember to treat extrapolation properly (e.g. replace with NaN?).

    if x is None:
        x = df.columns[0]
    if y is None:
        y = df.columns[1]

    xs = df[x].values
    ys = df[y].values

    if direction > 0:
        x_min = xs.min()
        x_max = xs.max()
    else:
        x_max = xs.min()
        x_min = xs.max()
        dx = -dx

    bounds_error = kwargs.pop("bounds_error", False)
    f = interpolate.interp1d(xs, ys, bounds_error=bounds_error, **kwargs)
    if new_x is None:
        if number_of_points:
            new_x = np.linspace(x_min, x_max, number_of_points)
        else:
            new_x = np.arange(x_min, x_max, dx)

    new_y = f(new_x)

    new_df = pd.DataFrame({x: new_x, y: new_y})

    return new_df


def group_by_interpolate(
    df,
    x=None,
    y=None,
    group_by=None,
    number_of_points=100,
    tidy=False,
    individual_x_cols=False,
    header_name="Unit",
    dx=10.0,
    generate_new_x=True,
):
    """Do a pandas.DataFrame.group_by and perform interpolation for all groups.

    This function is a wrapper around an internal interpolation function in
    cellpy (that uses scipy.interpolate.interp1d) that combines doing a group-by
    operation and interpolation.

    Args:
        df (pandas.DataFrame): the dataframe to morph.
        x (str): the header for the x-value
            (defaults to normal header step_time_txt) (remark that the default
            group_by column is the cycle column, and each cycle normally
            consist of several steps (so you risk interpolating / merging
            several curves on top of each other (not good)).
        y (str): the header for the y-value
            (defaults to normal header voltage_txt).
        group_by (str): the header to group by
            (defaults to normal header cycle_index_txt)
        number_of_points (int): if generating new x-column, how many values it
            should contain.
        tidy (bool): return the result in tidy (i.e. long) format.
        individual_x_cols (bool): return as xy xy xy ... data.
        header_name (str): name for the second level of the columns (only
            applies for xy xy xy ... data) (defaults to "Unit").
        dx (float): if generating new x-column and number_of_points is None or
            zero, distance between the generated values.
        generate_new_x (bool): create a new x-column by
            using the x-min and x-max values from the original dataframe where
            the method is set by the number_of_points key-word:

            1)  if number_of_points is not None (default is 100):

                ```
                new_x = np.linspace(x_max, x_min, number_of_points)
                ```
            2)  else:
                ```
                new_x = np.arange(x_max, x_min, dx)
                ```


    Returns: pandas.DataFrame with interpolated x- and y-values. The returned
        dataframe is in tidy (long) format for tidy=True.

    """
    # TODO: @jepe - create more tests
    time_00 = time.time()
    if x is None:
        x = HEADERS_NORMAL.step_time_txt
    if y is None:
        y = HEADERS_NORMAL.voltage_txt
    if group_by is None:
        group_by = [HEADERS_NORMAL.cycle_index_txt]

    if not isinstance(group_by, (list, tuple)):
        group_by = [group_by]

    if not generate_new_x:
        # check if it makes sence
        if (not tidy) and (not individual_x_cols):
            logging.warning("Unlogical condition")
            generate_new_x = True

    new_x = None

    if generate_new_x:
        x_max = df[x].max()
        x_min = df[x].min()
        if number_of_points:
            new_x = np.linspace(x_max, x_min, number_of_points)
        else:
            new_x = np.arange(x_max, x_min, dx)

    new_dfs = []
    keys = []

    for name, group in df.groupby(group_by):
        keys.append(name)
        if not isinstance(name, (list, tuple)):
            name = [name]

        new_group = interpolate_y_on_x(
            group, x=x, y=y, new_x=new_x, number_of_points=number_of_points, dx=dx
        )

        if tidy or (not tidy and not individual_x_cols):
            for i, j in zip(group_by, name):
                new_group[i] = j
        new_dfs.append(new_group)

    if tidy:
        new_df = pd.concat(new_dfs)
    else:
        if individual_x_cols:
            new_df = pd.concat(new_dfs, axis=1, keys=keys)
            group_by.append(header_name)
            new_df.columns.names = group_by
        else:
            new_df = pd.concat(new_dfs)
            new_df = new_df.pivot(index=x, columns=group_by[0], values=y)

    time_01 = time.time() - time_00
    logging.debug(f"duration: {time_01} seconds")
    return new_df


if __name__ == "__main__":
    logging.getLogger().setLevel(logging.DEBUG)
    register_instruments()
    # find_all_instruments()



"""diagnostics contains routines for creating extracting key
features from a set of cells"""

import numpy as np
import pandas as pd

from cellpy import prms
from cellpy.exceptions import NullData


def first_cycle_irreversible_capacity():
    ...



"""Tools for getting some data to play with"""
import logging
import os
from pathlib import Path

import cellpy

logging.info("Ready to help you to get some data to play with.")
CURRENT_PATH = Path(os.path.dirname(os.path.realpath(__file__)))
RAW_PATH = CURRENT_PATH / "data" / "raw"
H5_PATH = CURRENT_PATH / "data"


def arbin_file(auto_summary=True, testing=False):
    """load an example data file (arbin).

    Args:
        auto_summary (bool): run make_summary automatically (defaults to True)
        testing (bool): run in test mode

    Returns:
        cellpy.CellpyData object with the arbin data loaded

    """
    file_path = RAW_PATH / "20160805_test001_45_cc_01.res"
    mass = 0.704
    return cellpy.get(
        filename=file_path, mass=mass, auto_summary=auto_summary, testing=testing
    )


def cellpy_file(testing=False):
    """load an example cellpy file.

    Args:
        testing (bool): run in test mode

    Returns:
        cellpy.CellpyData object with the arbin data loaded
    """

    file_path = H5_PATH / "20160805_test001_45_cc.h5"
    return cellpy.get(filename=file_path, testing=testing)


def cellpy_file_path():
    """Get the path to an example cellpy file"""

    return H5_PATH / "20160805_test001_45_cc.h5"


def arbin_file_path():
    """Get the path to an example arbin res file"""

    return RAW_PATH / "20160805_test001_45_cc_01.res"



"""only used for testing and development"""

# not updated yet
unit_labels = {
    "resistance": "Ohms",
    # not observed yet:
    "time": "s",
    "current": "A",
    "voltage": "V",
    "power": "W",
    "capacity": "Ah",
    "energy": "Wh",
    "temperature": "C",
}

# not observed yet
incremental_unit_labels = {
    "dv_dt": f"{unit_labels['voltage']}/{unit_labels['time']}",
    "dq_dv": f"{unit_labels['capacity']}/{unit_labels['voltage']}",
    "dv_dq": f"{unit_labels['voltage']}/{unit_labels['capacity']}",
}


normal_headers_renaming_dict = {
    "data_point_txt": "Rec",
    "cycle_index_txt": "Cycle P",
    "step_index_txt": "Step",
    "test_time_txt": "TestTime",
    "step_time_txt": "StepTime",
    "charge_capacity_txt": "Cap. [Ah]",
    "charge_energy_txt": "Ener. [Wh]",
    "current_txt": "Current [A]",
    "voltage_txt": "Voltage [V]",
    "datetime_txt": "DPT Time",
    # not observed yet:
    "ac_impedance_txt": f"ACImp/{unit_labels['resistance']}",
    "internal_resistance_txt": f"DCIR/{unit_labels['resistance']}",
    "sub_step_index_txt": f"Sub_Step_Index",  # new
    "sub_step_time_txt": f"Sub_Step_Time",  # new
    "discharge_capacity_txt": f"Discharge_Capacity({unit_labels['capacity']})",
    "discharge_energy_txt": f"Discharge_Energy({unit_labels['energy']})",
    "dv_dt_txt": f"dV/dt({incremental_unit_labels['dv_dt']})",  # TODO: include the new cols into internal settings
    "dq_dv_txt": f"dV/dt({incremental_unit_labels['dq_dv']})",  # TODO: include the new cols into internal settings
    "dv_dq_txt": f"dV/dt({incremental_unit_labels['dv_dq']})",  # TODO: include the new cols into internal settings
    "acr_txt": f"Internal_Resistance({unit_labels['resistance']})",  # TODO: include the new cols into internal settings
    "aci_phase_angle_txt": f"ACI_Phase_Angle",
    "ref_aci_phase_angle_txt": f"Reference_ACI_Phase_Angle",
    "ref_ac_impedance_txt": f"Reference_AC_Impedance",
    "is_fc_data_txt": f"Is_FC_Data",
    "test_id_txt": f"Test_ID",
    "ref_voltage_txt": f"Reference_Voltage({unit_labels['resistance']})",  # new
    "frequency_txt": f"Frequency",  # new
    "amplitude_txt": f"Amplitude",  # new
    "channel_id_txt": f"Channel_ID",  # new Arbin SQL Server
    "data_flag_txt": f"Data_Flags",  # new Arbin SQL Server
    "test_name_txt": f"Test_Name",  # new Arbin SQL Server
}

# not observed yet
not_implemented_in_cellpy_yet_renaming_dict = {
    f"Power({unit_labels['power']})": "power",
    f"ACR({unit_labels['resistance']})": "acr",
    f"dV/dt({incremental_unit_labels['dv_dt']})": "dv_dt",
    f"dQ/dV({incremental_unit_labels['dq_dv']})": "dq_dv",
    f"dV/dQ({incremental_unit_labels['dv_dq']})": "dv_dq",
}

columns_to_keep = [
    "Rec",
    "Cycle P",
    "Cycle C",
    "Step",
    "TestTime",
    "StepTime",
    "Cap. [Ah]",
    "Ener. [Wh]",
    "Current [A]",
    "Voltage [V]",
    "Md",
    "DPT Time",
]

states = {
    "column_name": "Md",
    "charge_keys": ["C"],
    "discharge_keys": ["D"],
    "rest_keys": ["R"],
}


raw_units = {"current": 1.0, "charge": 1.0, "mass": 0.001}

raw_limits = {
    "current_hard": 0.000_000_000_000_1,
    "current_soft": 0.000_01,
    "stable_current_hard": 2.0,
    "stable_current_soft": 4.0,
    "stable_voltage_hard": 2.0,
    "stable_voltage_soft": 4.0,
    "stable_charge_hard": 0.001,
    "stable_charge_soft": 5.0,
    "ir_change": 0.00001,
}

formatters = {
    "skiprows": 12,  # 12 for other file
    "sep": "\t",
    "header": 0,  # 0 for other file
    "encoding": "ISO-8859-1",  # options: "ISO-8859-1", "utf-8", "cp1252"
    "decimal": ",",
}

pre_processors = {
    "remove_empty_lines": True,
}

post_processors = {
    "split_capacity": True,
    "split_current": True,
    "set_index": True,
    "rename_headers": True,
    "set_cycle_number_not_zero": True,
    "convert_date_time_to_datetime": True,
    "convert_step_time_to_timedelta": True,
    "convert_test_time_to_timedelta": True,
}



"""Internal settings and definitions and functions for getting them."""

import collections


class HeaderDict(collections.UserDict):
    """Sub-classing dict to allow for tab-completion."""

    def __setitem__(self, key: str, value: str) -> None:
        if key == "data":
            raise KeyError("protected key")
        super().__setitem__(key, value)
        self.__dict__[key] = value


class ValuesDict(collections.UserDict):
    """Sub-classing dict to allow for tab-completion."""

    def __setitem__(self, key: str, value: float) -> None:
        if key == "data":
            raise KeyError("protected key")
        super().__setitem__(key, value)
        self.__dict__[key] = value


cellpy_units: ValuesDict = ValuesDict()
cellpy_limits: ValuesDict = ValuesDict()
headers_normal: HeaderDict = HeaderDict()
headers_summary: HeaderDict = HeaderDict()
headers_step_table: HeaderDict = HeaderDict()
headers_journal: HeaderDict = HeaderDict()

# cellpy attributes that should be loaded from cellpy-files:

ATTRS_CELLPYFILE = [
    "mass",
    "channel_index",
    "channel_number",
    "creator",
    "cycle_mode",
    "schedule_file_name",
    "start_datetime",
    "test_ID",
    "cell_no",
    "name",
    "nom_cap",
    "material",
    "item_ID",
    "active_electrode_area",
    "active_electrode_thickness",
    "electrolyte_type",
    "electrolyte_volume",
    "active_electrode_type",
    "counter_electrode_type",
    "reference_electrode_type",
    "experiment_type",
    "active_electrode_current_collector",
    "reference_electrode_current_collector",
    "comment",
]

# Attributes that should be copied when duplicating cellpy objects:

# current attributes for the cellpy.cellpydata objects
ATTRS_CELLPYDATA = [
    "auto_dirs",
    "capacity_modifiers",
    "cellpy_datadir",
    "daniel_number",
    "ensure_step_table",
    "file_names",
    "filestatuschecker",
    "force_all",
    "force_step_table_creation",
    "forced_errors",
    "limit_loaded_cycles",
    "load_only_summary",
    "minimum_selection",
    "name",
    "number_of_datasets",
    "profile",
    "raw_datadir",
    "raw_limits",
    "raw_units",
    "select_minimal",
    "selected_cell_number",
    "selected_scans",
    "sep",
    "status_datasets",
    "summary_exists",
    "table_names",
    "tester",
]

# current attributes used for the cellpy.cell objects
ATTRS_DATASET = [
    "cellpy_file_version",
    "channel_index",
    "channel_number",
    "charge_steps",
    "creator",
    "cycle_mode",
    # "data",
    "discharge_steps",
    "file_errors",
    "ir_steps",
    "item_ID",
    "loaded_from",
    "mass",
    "mass_given",
    "material",
    "merged",
    "name",
    "no_cycles",
    "nom_cap",
    "ocv_steps",
    "raw_data_files_length",
    "raw_limits",
    "raw_units",
    "schedule_file_name",
    "start_datetime",
    # "summary",
    "test_ID",
    "cell_no",
    "tot_mass",
]

ATTRS_DATASET_DEEP = ["raw_data_files"]

# cellpy units:

cellpy_units["current"]: float = 0.001  # mA
cellpy_units["charge"]: float = 0.001  # Ah
cellpy_units["mass"]: float = 0.001  # mg (used for input of mass)
cellpy_units["specific"]: float = 1.0  # g (used for calc. of e.g. spec. capacity)
cellpy_units["voltage"]: float = 1.0  # V (not implemented yet)


# cellpy limits:

cellpy_limits["current_hard"]: float = 0.0000000000001
cellpy_limits["current_soft"]: float = 0.00001
cellpy_limits["stable_current_hard"]: float = 2.0
cellpy_limits["stable_current_soft"]: float = 4.0
cellpy_limits["stable_voltage_hard"]: float = 2.0
cellpy_limits["stable_voltage_soft"]: float = 4.0
cellpy_limits["stable_charge_hard"]: float = 0.9
cellpy_limits["stable_charge_soft"]: float = 5.0
cellpy_limits["ir_change"]: float = 0.00001


# headers for out-files:

# 01.05.2020: renamed the column names to the form cycle_index, discharge_capacity_u_mAh_g, etc.

# - normal (data) -

headers_normal["aci_phase_angle_txt"] = "aci_phase_angle"
headers_normal["ref_aci_phase_angle_txt"] = "ref_aci_phase_angle"
headers_normal["ac_impedance_txt"] = "ac_impedance"
headers_normal["ref_ac_impedance_txt"] = "ref_ac_impedance"  # new
headers_normal["charge_capacity_txt"] = "charge_capacity"
headers_normal["charge_energy_txt"] = "charge_energy"
headers_normal["current_txt"] = "current"
headers_normal["cycle_index_txt"] = "cycle_index"
headers_normal["data_point_txt"] = "data_point"
headers_normal["datetime_txt"] = "date_time"
headers_normal["discharge_capacity_txt"] = "discharge_capacity"
headers_normal["discharge_energy_txt"] = "discharge_energy"
headers_normal["internal_resistance_txt"] = "internal_resistance"
headers_normal["is_fc_data_txt"] = "is_fc_data"
headers_normal["step_index_txt"] = "step_index"
headers_normal["sub_step_index_txt"] = "sub_step_index"  # new
headers_normal["step_time_txt"] = "step_time"
headers_normal["sub_step_time_txt"] = "sub_step_time"  # new
headers_normal["test_id_txt"] = "test_id"
headers_normal["test_time_txt"] = "test_time"
headers_normal["voltage_txt"] = "voltage"
headers_normal["ref_voltage_txt"] = "reference_voltage"  # new
headers_normal["dv_dt_txt"] = "dv_dt"
headers_normal["frequency_txt"] = "frequency"  # new
headers_normal["amplitude_txt"] = "amplitude"  # new
headers_normal["channel_id_txt"] = "channel_id"  # new Arbin SQL Server
headers_normal["data_flag_txt"] = "data_flag"  # new Arbin SQL Server
headers_normal["test_name_txt"] = "test_name"  # new Arbin SQL Server
headers_normal["test_id_txt"] = "test_id"  # new Arbin SQL Server
# - summary -

# 08.12.2016: added temperature_last, temperature_mean, aux_
headers_summary["cycle_index"] = headers_normal["cycle_index_txt"]
headers_summary["data_point"] = headers_normal["data_point_txt"]
headers_summary["test_time"] = headers_normal["test_time_txt"]
headers_summary["datetime"] = headers_normal["datetime_txt"]
headers_summary["discharge_capacity_raw"] = headers_normal["discharge_capacity_txt"]
headers_summary["charge_capacity_raw"] = headers_normal["charge_capacity_txt"]
headers_summary["test_name"] = headers_normal["test_name_txt"]
headers_summary["data_flag"] = headers_normal["data_flag_txt"]
headers_summary["channel_id"] = headers_normal["channel_id_txt"]
headers_summary["discharge_capacity"] = "discharge_capacity_u_mAh_g"
headers_summary["charge_capacity"] = "charge_capacity_u_mAh_g"
headers_summary["cumulated_charge_capacity"] = "cumulated_charge_capacity_u_mAh_g"
headers_summary["cumulated_discharge_capacity"] = "cumulated_discharge_capacity_u_mAh_g"
headers_summary["coulombic_efficiency"] = "coulombic_efficiency_u_percentage"
headers_summary[
    "cumulated_coulombic_efficiency"
] = "cumulated_coulombic_efficiency_u_percentage"
headers_summary["coulombic_difference"] = "coulombic_difference_u_mAh_g"
headers_summary[
    "cumulated_coulombic_difference"
] = "cumulated_coulombic_difference_u_mAh_g"
headers_summary["discharge_capacity_loss"] = "discharge_capacity_loss_u_mAh_g"
headers_summary["charge_capacity_loss"] = "charge_capacity_loss_u_mAh_g"
headers_summary[
    "cumulated_discharge_capacity_loss"
] = "cumulated_discharge_capacity_loss_u_mAh_g"
headers_summary[
    "cumulated_charge_capacity_loss"
] = "cumulated_charge_capacity_loss_u_mAh_g"
headers_summary["ir_discharge"] = "ir_discharge_u_Ohms"
headers_summary["ir_charge"] = "ir_charge_u_Ohms"
headers_summary["ocv_first_min"] = "ocv_first_min_u_V"
headers_summary["ocv_second_min"] = "ocv_second_min_u_V"
headers_summary["ocv_first_max"] = "ocv_first_max_u_V"
headers_summary["ocv_second_max"] = "ocv_second_max_u_V"
headers_summary["end_voltage_discharge"] = "end_voltage_discharge_u_V"
headers_summary["end_voltage_charge"] = "end_voltage_charge_u_V"
headers_summary["cumulated_ric_disconnect"] = "cumulated_ric_disconnect_u_none"
headers_summary["cumulated_ric_sei"] = "cumulated_ric_sei_u_none"
headers_summary["cumulated_ric"] = "cumulated_ric_u_none"

headers_summary["normalized_cycle_index"] = "normalized_cycle_index"
headers_summary["normalized_charge_capacity"] = "normalized_charge_capacity"
headers_summary["normalized_discharge_capacity"] = "normalized_discharge_capacity"

# Sum of irreversible capacity:
headers_summary["low_level"] = "low_level_u_percentage"
# SEI loss:
headers_summary["high_level"] = "high_level_u_percentage"
# Shifted capacities:
headers_summary["shifted_charge_capacity"] = "shifted_charge_capacity_u_mAh_g"
headers_summary["shifted_discharge_capacity"] = "shifted_discharge_capacity_u_mAh_g"
# Other
headers_summary["temperature_last"] = "temperature_last_u_C"
headers_summary["temperature_mean"] = "temperature_mean_u_C"
headers_summary["areal_charge_capacity"] = "areal_charge_capacity_u_mAh_cm2"
headers_summary["areal_discharge_capacity"] = "areal_discharge_capacity_u_mAh_cm2"
headers_summary["charge_c_rate"] = "charge_c_rate"
headers_summary["discharge_c_rate"] = "discharge_c_rate"
headers_summary["pre_aux"] = "aux_"

# - step table -

# 08.12.2016: added sub_step, sub_type, and pre_time

headers_step_table["test"] = "test"
headers_step_table["ustep"] = "ustep"
headers_step_table["cycle"] = "cycle"
headers_step_table["step"] = "step"
headers_step_table["test_time"] = "test_time"
headers_step_table["step_time"] = "step_time"
headers_step_table["sub_step"] = "sub_step"
headers_step_table["type"] = "type"
headers_step_table["sub_type"] = "sub_type"
headers_step_table["info"] = "info"
headers_step_table["voltage"] = "voltage"
headers_step_table["current"] = "current"
headers_step_table["charge"] = "charge"
headers_step_table["discharge"] = "discharge"
headers_step_table["point"] = "point"
headers_step_table["internal_resistance"] = "ir"
headers_step_table["internal_resistance_change"] = "ir_pct_change"
headers_step_table["rate_avr"] = "rate_avr"

# 01.05.2020: added fix column names and renamed to singular form.
# 08.07.2020: added nominal capacity, experiment type and comment.
# 10.08.2021: added id_key (can be used for look-up in other dbs).
headers_journal["filename"] = "filename"
headers_journal["id_key"] = "id_key"
headers_journal["mass"] = "mass"
headers_journal["total_mass"] = "total_mass"
headers_journal["loading"] = "loading"
headers_journal["nom_cap"] = "nom_cap"
headers_journal["experiment"] = "experiment"
headers_journal["fixed"] = "fixed"
headers_journal["label"] = "label"
headers_journal["cell_type"] = "cell_type"
headers_journal["instrument"] = "instrument"
headers_journal["raw_file_names"] = "raw_file_names"
headers_journal["cellpy_file_name"] = "cellpy_file_name"
headers_journal["group"] = "group"
headers_journal["sub_group"] = "sub_group"
headers_journal["comment"] = "comment"

keys_journal_session = ["starred", "bad_cells", "bad_cycles", "notes"]


def get_headers_summary() -> HeaderDict:
    """Returns a dictionary containing the header-strings for the summary
    (used as column headers for the summary pandas DataFrames)"""
    # maybe I can do some tricks in here so that tab completion works?
    # ctrl + space works
    return headers_summary


def get_cellpy_units() -> ValuesDict:
    """Returns a dictionary with units"""
    return cellpy_units


def get_headers_normal() -> HeaderDict:
    """Returns a dictionary containing the header-strings for the normal data
    (used as column headers for the main data pandas DataFrames)"""
    return headers_normal


def get_headers_step_table() -> HeaderDict:
    """Returns a dictionary containing the header-strings for the steps table
    (used as column headers for the steps pandas DataFrames)"""
    return headers_step_table


def get_headers_journal() -> HeaderDict:
    """Returns a dictionary containing the header-strings for the journal (batch)
    (used as column headers for the journal pandas DataFrames)"""
    return headers_journal



"""Engines are functions that are used by the Do-ers.

    Keyword Args: experiments, farms, barn, optionals
    Returns: farms, barn
"""

import logging
import time

import pandas as pd

from cellpy import dbreader
from cellpy.parameters.internal_settings import headers_journal
from cellpy.utils.batch_tools import batch_helpers as helper

# logger = logging.getLogger(__name__)

SELECTED_SUMMARIES = [
    "discharge_capacity",
    "charge_capacity",
    "coulombic_efficiency",
    "cumulated_coulombic_efficiency",
    "ir_discharge",
    "ir_charge",
    "end_voltage_discharge",
    "end_voltage_charge",
    "charge_c_rate",
    "discharge_c_rate",
]


def cycles_engine(**kwargs):
    """engine to extract cycles"""
    logging.debug("cycles_engine::Not finished yet (sorry).")
    # raise NotImplementedError

    experiments = kwargs["experiments"]

    farms = []
    barn = "raw_dir"  # Its a murder in the red barn - murder in the red barn

    for experiment in experiments:
        farms.append([])
        if experiment.all_in_memory:
            logging.debug("all in memory")
            for key in experiment.cell_data_frames:
                logging.debug(f"extracting cycles from {key}")
                # extract cycles here and send it to the farm
        else:
            logging.debug("dont have it in memory - need to lookup in the files")
            for key in experiment.cell_data_frames:
                logging.debug(f"looking up cellpyfile for {key}")
                # extract cycles here and send it to the farm

    return farms, barn


def raw_data_engine(**kwargs):
    """engine to extract raw data"""
    logging.debug("cycles_engine")
    farms = None
    barn = "raw_dir"
    raise NotImplementedError


def summary_engine(**kwargs):
    """engine to extract summary data"""
    logging.debug("summary_engine")
    # farms = kwargs["farms"]

    farms = []
    experiments = kwargs.pop("experiments")
    reset = kwargs.pop("reset", False)

    for experiment in experiments:
        if experiment.selected_summaries is None:
            selected_summaries = SELECTED_SUMMARIES
        else:
            selected_summaries = experiment.selected_summaries

        if reset or experiment.summary_frames is None:
            logging.debug("No summary frames found")
            logging.debug("Re-loading")
            experiment.summary_frames = _load_summaries(experiment)

        farm = helper.join_summaries(experiment.summary_frames, selected_summaries)
        farms.append(farm)
    barn = "batch_dir"

    return farms, barn


def _load_summaries(experiment):
    summary_frames = {}
    for label in experiment.cell_names:
        # TODO: replace this with direct lookup from hdf5?
        summary_frames[label] = experiment.data[label].cell.summary
    return summary_frames


def dq_dv_engine(**kwargs):
    """engine that performs incremental analysis of the cycle-data"""
    farms = None
    barn = "raw_dir"
    raise NotImplementedError


def _query(reader_method, cell_ids, column_name=None):
    if not any(cell_ids):
        logging.debug("Received empty cell_ids")
        return []

    try:
        if column_name is None:
            result = [reader_method(cell_id) for cell_id in cell_ids]
        else:
            result = [reader_method(column_name, cell_id) for cell_id in cell_ids]
    except Exception as e:
        logging.debug(f"Error in querying db.")
        logging.debug(e)
        result = [None for _ in range(len(cell_ids))]
    return result


def simple_db_engine(
    reader=None,
    cell_ids=None,
    file_list=None,
    pre_path=None,
    include_key=False,
    include_individual_arguments=True,
    additional_column_names=None,
    **kwargs,
):
    """Engine that gets values from the db for given set of cell IDs.

    The simple_db_engine looks up values for mass, names, etc. from
    the db using the reader object. In addition, it searches for the
    corresponding raw files / data.

    Args:
        reader: a reader object (defaults to dbreader.Reader)
        cell_ids: keys (cell IDs)
        file_list: file list to send to filefinder (instead of searching in folders for files).
        pre_path: prepended path to send to filefinder.
        include_key: include the key col in the pages (the cell IDs).
        include_individual_arguments: include the argument column in the pages.
        additional_column_names: list of additional column names to include in the pages.
        **kwargs: sent to filefinder

    Returns:
        pages (pandas.DataFrame)
    """

    # This is not really a proper Do-er engine. But not sure where to put it.
    if reader is None:
        reader = dbreader.Reader()
        logging.debug("No reader provided. Creating one myself.")
    pages_dict = dict()
    pages_dict[headers_journal["filename"]] = _query(reader.get_cell_name, cell_ids)
    if include_key:
        pages_dict[headers_journal["id_key"]] = cell_ids

    if include_individual_arguments:
        pages_dict[headers_journal["argument"]] = _query(reader.get_args, cell_ids)

    pages_dict[headers_journal["mass"]] = _query(reader.get_mass, cell_ids)
    pages_dict[headers_journal["total_mass"]] = _query(reader.get_total_mass, cell_ids)
    pages_dict[headers_journal["loading"]] = _query(reader.get_loading, cell_ids)
    pages_dict[headers_journal["nom_cap"]] = _query(reader.get_nom_cap, cell_ids)
    pages_dict[headers_journal["experiment"]] = _query(
        reader.get_experiment_type, cell_ids
    )
    pages_dict[headers_journal["fixed"]] = _query(reader.inspect_hd5f_fixed, cell_ids)
    pages_dict[headers_journal["label"]] = _query(reader.get_label, cell_ids)
    pages_dict[headers_journal["cell_type"]] = _query(reader.get_cell_type, cell_ids)
    pages_dict[headers_journal["instrument"]] = _query(reader.get_instrument, cell_ids)
    pages_dict[headers_journal["raw_file_names"]] = []
    pages_dict[headers_journal["cellpy_file_name"]] = []
    pages_dict[headers_journal["comment"]] = _query(reader.get_comment, cell_ids)

    if additional_column_names is not None:
        for k in additional_column_names:
            try:
                pages_dict[k] = _query(reader.get_by_column_label, cell_ids, k)
            except Exception as e:
                logging.info(f"Could not retrieve from column {k} ({e})")

    # get id_key (not implemented yet

    logging.debug(f"created info-dict from {reader.db_file}:")
    # logging.debug(info_dict)

    for key in list(pages_dict.keys()):
        logging.debug("%s: %s" % (key, str(pages_dict[key])))

    _groups = _query(reader.get_group, cell_ids)

    logging.debug(">\ngroups: %s" % str(_groups))
    groups = helper.fix_groups(_groups)
    pages_dict[headers_journal["group"]] = groups

    my_timer_start = time.time()
    pages_dict = helper.find_files(
        pages_dict, file_list=file_list, pre_path=pre_path, **kwargs
    )
    my_timer_end = time.time()
    if (my_timer_end - my_timer_start) > 5.0:
        logging.critical(
            "The function _find_files was very slow. "
            "Save your journal so you don't have to run it again! "
            "You can load it again using the from_journal(journal_name) method."
        )

    pages = pd.DataFrame(pages_dict)
    try:
        pages = pages.sort_values([headers_journal.group, headers_journal.filename])
    except TypeError as e:
        _report_suspected_duplicate_id(
            e,
            "sort the values",
            pages[[headers_journal.group, headers_journal.filename]],
        )

    pages = helper.make_unique_groups(pages)

    try:
        pages[headers_journal.label] = pages[headers_journal.filename].apply(
            helper.create_labels
        )
    except AttributeError as e:
        _report_suspected_duplicate_id(
            e, "make labels", pages[[headers_journal.label, headers_journal.filename]]
        )

    else:
        # TODO: check if drop=False works [#index]
        pages.set_index(
            headers_journal["filename"], inplace=True
        )  # edit this to allow for
        # non-numeric index-names (for tab completion and python-box)
    return pages


def _report_suspected_duplicate_id(e, what="do it", on=None):
    logging.warning(f"could not {what}")
    logging.warning(f"{on}")
    logging.warning("maybe you have a corrupted db?")
    logging.warning(
        "typically happens if the cell_id is not unique (several rows or records in "
        "your db has the same cell_id or key)"
    )
    logging.warning(e)



from cellpy.utils.batch_tools.batch_core import BaseReporter


class HTMLReporter(BaseReporter):
    def __init__(self):
        super().__init__()


class PPTReporter(BaseReporter):
    def __init__(self):
        super().__init__()



"""Set up logger instance"""

import datetime
import json
import logging
import logging.config
import os
import pathlib
import shutil
import tempfile
import warnings

from cellpy import prms

logging.raiseExceptions = False


def setup_logging(
    default_json_path=None,
    default_level=None,
    env_key="LOG_CFG",
    custom_log_dir=None,
    reset_big_log=False,
    max_size=5_000_000,
    testing=False,
):
    """Setup logging configuration.

    Args:
        default_level: default log-level to screen (std.out).
        default_json_path: path to config file for setting up logging.
        env_key (str): use this environment prm to try to get default_json_path.
        custom_log_dir: path for saving logs.
        reset_big_log (bool): reset log if too big (max_size).
        max_size (int): if reset_log, this is the max limit.
        testing (bool): set as True if testing, and you don't want to create any .log files

    """

    if not default_json_path:
        default_json_path = os.path.join(
            os.path.dirname(os.path.realpath(__file__)), "logging.json"
        )
    path = default_json_path
    value = os.getenv(env_key, None)
    if value:
        path = value

    if default_level is None:
        default_level = "CRITICAL"

    # loading logging configs
    if os.path.exists(path):
        with open(path, "rt") as f:
            config = json.load(f)

        if testing:
            log_dir = tempfile.mkdtemp()
        elif custom_log_dir:
            log_dir = custom_log_dir
        else:
            log_dir = os.path.abspath(prms.Paths.filelogdir)

        if not os.path.isdir(log_dir):
            warning_txt = (
                "\nCould not set custom log-dir - "
                "non-existing directory"
                f"\nDir: {log_dir}"
                "\nUsing current directory instead: "
                f"{os.getcwd()}"
            )
            logging.warning(warning_txt)
            log_dir = os.getcwd()

        for file_handler in [
            "error_file_handler",
            "info_file_handler",
            "debug_file_handler",
        ]:
            try:
                file_name = config["handlers"][file_handler]["filename"]
                logging.debug("Setting file handlers for logging.")
                logging.debug(f"Filename: {file_name}")
                logging.debug(f"Full path: {os.path.join(log_dir,file_name)}")
                # print(f"Filename: {file_name}")
                # print(f"Full path: {os.path.join(log_dir,file_name)}")
                config["handlers"][file_handler]["filename"] = os.path.join(
                    log_dir, file_name
                )

                if reset_big_log:
                    full_log_file_path = pathlib.Path(log_dir) / file_name
                    if full_log_file_path.is_file():
                        file_size = full_log_file_path.lstat().st_size
                        if file_size > max_size:
                            d_str = datetime.datetime.now().strftime("%Y_%m_%d_%H_%M")
                            new_file_name = "_".join([d_str, file_name])
                            new_full_log_file_path = (
                                pathlib.Path(log_dir) / new_file_name
                            )
                            shutil.copy(full_log_file_path, new_full_log_file_path)
                    else:
                        logging.debug(
                            "Could not reset big log: could not find the file"
                        )

            except Exception as e:
                warnings.warn("\nCould not set custom log-dir" + str(e))

        if default_level:
            w_txt = "\nCould not set custom default level for logger"
            if default_level not in [
                "INFO",
                "DEBUG",
                "CRITICAL",
                logging.INFO,
                logging.DEBUG,
                logging.CRITICAL,
            ]:
                _txt = "\nonly 'INFO', 'DEBUG' and 'CRITICAL' is supported"
                _txt += " as default_level"
                warnings.warn(w_txt + _txt)

            else:
                try:
                    config["handlers"]["console"]["level"] = default_level
                    if default_level in ["DEBUG", logging.DEBUG]:
                        config["handlers"]["console"]["formatter"] = "stamped"

                except Exception as e:
                    warnings.warn(w_txt + "\n" + str(e))

        logging.config.dictConfig(config)
    else:
        if not default_level:
            default_level = logging.CRITICAL
        logging.basicConfig(level=default_level)


if __name__ == "__main__":
    setup_logging()



"""Dumpers are functions that are used by the Do-ers.
Keyword Args: experiments, farms, barn, engine
Returns nothing.
"""

import logging
import os

from cellpy import prms
from cellpy.utils.batch_tools.batch_helpers import generate_folder_names


def csv_dumper(**kwargs):
    """dump data to csv"""
    logging.debug("dumping to csv")
    barn = kwargs["barn"]
    farms = kwargs["farms"]
    experiments = kwargs["experiments"]
    for experiment, farm in zip(experiments, farms):
        name = experiment.journal.name
        project = experiment.journal.project
        project_dir, batch_dir, raw_dir = experiment.journal.paginate()
        if batch_dir is None:
            logging.debug("have to generate folder-name on the fly")
            out_data_dir, project_dir, batch_dir, raw_dir = generate_folder_names(
                name, project
            )

        if barn == "batch_dir":
            out_dir = batch_dir
        elif barn == "project_dir":
            out_dir = project_dir
        elif barn == "raw_dir":
            out_dir = raw_dir
        else:
            out_dir = barn

        for animal in farm:
            file_name = os.path.join(out_dir, "summary_%s_%s.csv" % (animal.name, name))
            logging.info(f"> {file_name}")
            animal.to_csv(file_name, sep=prms.Reader.sep)


def excel_dumper(**kwargs):
    """Dump data to excel xlxs-format."""
    pass


def origin_dumper(**kwargs):
    """Dump data to a format suitable for use in OriginLab."""
    pass


def ram_dumper(**kwargs):
    """Dump data to 'memory' for later usage."""
    logging.debug("trying to save stuff in memory")
    farms = kwargs["farms"]
    experiments = kwargs["experiments"]
    engine = kwargs["engine"]

    try:
        engine_name = engine.__name__
    except AttributeError:
        engine_name = engine.__dict__.__name__

    accepted_engines = ["summary_engine"]
    if engine_name in accepted_engines:
        logging.debug(
            "found the engine that I will try to dump from: " f"{engine_name}"
        )

        for experiment, farm in zip(experiments, farms):
            name = experiment.journal.name
            project = experiment.journal.project
            experiment.memory_dumped[engine_name] = farm
            logging.debug(f"farm put into memory_dumped ({project}::{name})")


def screen_dumper(**kwargs):
    """Dump data to screen."""
    farms = kwargs["farms"]
    engine = kwargs["engine"]
    logging.info("dumping to screen")

    print(f"\n[Screen dumper] ({engine})")
    try:
        if len(farms) == 1:
            print(f"You have one farm with little pandas.")

        else:
            print(f"You have {len(farms)} farms with little pandas.")
    except TypeError:
        print(" - your farm has burned to the ground.")
    else:
        for number, farm in enumerate(farms):
            print(f"[#{number + 1}]You have {len(farm)} little pandas in this farm.")
            for animal in farm:
                print(80 * "=")
                try:
                    print(animal.name)
                except AttributeError:
                    print("no-name")
                print(80 * "-")
                print(animal.head(5))
                print()



# -*- coding: utf-8 -*-

"""

"""

__author__ = "Jan Petter Maehlen", "Asbjoern Ulvestad", "Muhammad Abdelhamid"
__email__ = "jepe@ife.no", "Asbjorn.Ulvestad@ife.no", "Muhammad.Abdelhamid@ife.no"

import logging
import warnings

import cellpy._version
from cellpy.parameters import prms  # TODO: this might give circular ref
from cellpy.parameters import prmreader
from cellpy.readers import cellreader, dbreader, filefinder

__version__ = cellpy._version.__version__

logging.getLogger(__name__).addHandler(logging.NullHandler())

try:
    prmreader._read_prm_file(prmreader._get_prm_file())
except FileNotFoundError:
    warnings.warn("Could not find the config-file")
except UserWarning:
    warnings.warn("Could not read the config-file")

get = cellreader.get
__all__ = ["cellreader", "dbreader", "prmreader", "prms", "filefinder", "get"]



"""Maccor txt data"""

import pandas as pd

from cellpy import prms
from cellpy import exceptions
from cellpy.parameters.internal_settings import (
    HeaderDict,
    base_columns_float,
    base_columns_int,
    headers_normal,
)
from cellpy.readers.instruments.base import TxtLoader


SUPPORTED_MODELS = {
    "ZERO": "maccor_txt_zero",
    "ONE": "maccor_txt_one",
    "TWO": "maccor_txt_two",
    "THREE": "maccor_txt_three",
    "FOUR": "maccor_txt_four",
    "WMG_SIMBA": "maccor_txt_three",
    "KIT_SIMBA": "maccor_txt_four",
    "KIT_COMMA_SIMBA": "maccor_txt_two",
    "UBHAM_SIMBA": "maccor_txt_three",
    "S4000-WMG": "maccor_txt_four",
    "S4000-KIT": "maccor_txt_four",
    "S4000-UBHAM": "maccor_txt_three",
    }


MUST_HAVE_RAW_COLUMNS = [
    headers_normal.test_time_txt,
    headers_normal.step_time_txt,
    headers_normal.current_txt,
    headers_normal.voltage_txt,
    headers_normal.step_index_txt,
    headers_normal.cycle_index_txt,
    headers_normal.charge_capacity_txt,
    headers_normal.discharge_capacity_txt,
]


class MaccorTxtLoader(TxtLoader):
    """Class for loading data from Maccor txt files."""

    name = "maccor_txt"
    raw_ext = "txt"

    default_model = prms.Instruments.Maccor["default_model"]  # Required
    supported_models = SUPPORTED_MODELS  # Required

    @staticmethod
    def get_headers_aux(raw):
        """Defines the so-called auxiliary table column headings"""

        headers = HeaderDict()
        for col in raw.columns:
            if col.startswith("Aux_"):
                ncol = col.replace("/", "_")
                ncol = "".join(ncol.split("(")[0])
                headers[col] = ncol.lower()

        return headers

    def validate(self, data):
        """A simple check that all the needed columns has been successfully
        loaded and that they get the correct type"""
        missing_must_have_columns = []

        # validating the float-type raw data
        for col in base_columns_float:
            if col in data.raw.columns:
                data.raw[col] = pd.to_numeric(data.raw[col], errors="coerce")
            else:
                if col in MUST_HAVE_RAW_COLUMNS:
                    missing_must_have_columns.append(col)

        # validating the integer-type raw data
        for col in base_columns_int:
            if col in data.raw.columns:
                data.raw[col] = pd.to_numeric(
                    data.raw[col], errors="coerce", downcast="integer"
                )
            else:
                if col in MUST_HAVE_RAW_COLUMNS:
                    missing_must_have_columns.append(col)

        if missing_must_have_columns:
            raise exceptions.IOError(
                f"Missing needed columns: {missing_must_have_columns}\nAborting!"
            )
        return data


def check_retrieve_file(n=1):
    import pathlib

    pd.options.display.max_columns = 100
    # prms.Reader.sep = "\t"
    data_root = pathlib.Path(r"C:\scripting\cellpy_dev_resources")
    data_dir = data_root / r"2021_leafs_data\Charge-Discharge\Maccor series 4000"
    if n == 2:
        name = data_dir / "KIT-Full-cell-PW-HC-CT-cell016.txt"
    else:
        name = data_dir / "01_UBham_M50_Validation_0deg_01.txt"
    print(name)
    print(f"Exists? {name.is_file()}")
    if name.is_file():
        return name
    else:
        raise IOError(f"could not locate the file {name}")


def check_dev_loader(name=None, model=None):
    if name is None:
        name = check_retrieve_file()

    pd.options.display.max_columns = 100
    # prms.Reader.sep = "\t"

    sep = "\t"
    loader1 = MaccorTxtLoader(sep=sep, model=model)
    loader2 = MaccorTxtLoader(model="one")
    loader3 = MaccorTxtLoader(model="zero")
    loader4 = MaccorTxtLoader(model="zero")
    dd = loader1.loader(name)
    dd = loader2.loader(name)
    dd = loader3.loader(name)
    dd = loader4.loader(name)
    raw = dd[0].raw
    print(len(raw))


def check_dev_loader2(name=None, model=None, sep=None, number=2):
    if name is None:
        name = check_retrieve_file(number)

    pd.options.display.max_columns = 100

    if sep is not None and sep != "none":
        loader3 = MaccorTxtLoader(sep=sep, model=model)
    elif sep == "none":
        loader3 = MaccorTxtLoader(sep=None, model=model)
    else:
        loader3 = MaccorTxtLoader(model=model)

    dd = loader3.loader(name)

    raw = dd[0].raw
    print(len(raw))
    print(raw)


def check_loader(name=None, number=1, model="one"):
    import matplotlib.pyplot as plt

    if name is None:
        name = check_retrieve_file(number)
    print(name)
    pd.options.display.max_columns = 100
    # prms.Reader.sep = "\t"

    loader = MaccorTxtLoader(sep="\t", model=model)
    dd = loader.loader(name)
    raw = dd[0].raw
    raw.plot(x="data_point", y="current", title="current vs data-point")
    raw.plot(
        x="data_point",
        y=["charge_capacity", "discharge_capacity"],
        title="capacity vs data-point",
    )
    raw.plot(
        x="test_time",
        y=["charge_capacity", "discharge_capacity"],
        title="capacity vs test-time",
    )
    raw.plot(
        x="step_time",
        y=["charge_capacity", "discharge_capacity"],
        title="capacity vs step-time",
    )
    print(raw.head())
    plt.show()


def check_loader_from_outside():
    # NOT EDITED YET!!!
    import pathlib

    import matplotlib.pyplot as plt

    from cellpy import cellreader

    pd.options.display.max_columns = 100
    datadir = pathlib.Path(
        r"C:\scripts\cellpy_dev_resources\2021_leafs_data\Charge-Discharge\Maccor series 4000"
    )
    name = datadir / "01_UBham_M50_Validation_0deg_01.txt"
    out = pathlib.Path(r"C:\scripts\notebooks\Div")
    print(f"Exists? {name.is_file()}")

    c = cellreader.CellpyData()
    c.set_instrument("maccor_txt", sep="\t")

    c.from_raw(name)
    c.set_mass(1000)

    c.make_step_table()
    c.make_summary()

    raw = c.cell.raw
    steps = c.cell.steps
    summary = c.cell.summary
    raw.to_csv(r"C:\scripts\notebooks\Div\trash\raw.csv", sep=";")
    steps.to_csv(r"C:\scripts\notebooks\Div\trash\steps.csv", sep=";")
    summary.to_csv(r"C:\scripts\notebooks\Div\trash\summary.csv", sep=";")

    fig_1, (ax1, ax2, ax3) = plt.subplots(3, 1, figsize=(6, 10))
    raw.plot(x="test_time", y="voltage", ax=ax1)
    raw.plot(x="test_time", y=["charge_capacity", "discharge_capacity"], ax=ax3)
    raw.plot(x="test_time", y="current", ax=ax2)

    n = c.get_number_of_cycles()
    print(f"number of cycles: {n}")

    cycle = c.get_cap(1, method="forth-and-forth")
    print(cycle.head())

    fig_2, (ax4, ax5, ax6) = plt.subplots(1, 3)
    cycle.plot(x="capacity", y="voltage", ax=ax4)
    s = c.get_step_numbers()
    t = c.sget_timestamp(1, s[1])
    v = c.sget_voltage(1, s[1])
    steps = c.sget_step_numbers(1, s[1])

    print("step numbers:")
    print(s)
    print("sget step numbers:")
    print(steps)
    print("\ntesttime:")
    print(t)
    print("\nvoltage")
    print(v)

    ax5.plot(t, v, label="voltage")
    ax6.plot(t, steps, label="steps")

    fig_3, (ax7, ax8) = plt.subplots(2, sharex=True)
    raw.plot(x="test_time", y="voltage", ax=ax7)
    raw.plot(x="test_time", y="step_index", ax=ax8)

    plt.legend()
    plt.show()

    outfile = out / "test_out"
    c.save(outfile)


def check_loader_from_outside_with_get():
    import pathlib

    import matplotlib.pyplot as plt

    import cellpy

    pd.options.display.max_columns = 100
    datadir = pathlib.Path(
        r"C:\scripting\cellpy_dev_resources\2021_leafs_data\Charge-Discharge\Maccor series 4000"
    )
    name = datadir / "01_UBham_M50_Validation_0deg_01.txt"
    out = pathlib.Path(r"C:\scripting\trash")
    print(f"File exists? {name.is_file()}")
    if not name.is_file():
        print(f"could not find {name} ")
        return

    c = cellpy.get(filename=name, instrument="maccor_txt", model="one", mass=1.0)
    print("loaded")
    raw = c.cell.raw
    steps = c.cell.steps
    summary = c.cell.summary

    raw.to_csv(r"C:\scripting\trash\raw.csv", sep=";")
    steps.to_csv(r"C:\scripting\trash\steps.csv", sep=";")
    summary.to_csv(r"C:\scripting\trash\summary.csv", sep=";")

    fig_1, (ax1, ax2, ax3) = plt.subplots(3, 1, figsize=(6, 10))
    raw.plot(x="test_time", y="voltage", ax=ax1, title="voltage")
    raw.plot(
        x="test_time", y=["charge_capacity", "discharge_capacity"], ax=ax3, title="caps"
    )
    raw.plot(x="test_time", y="current", ax=ax2, title="current")

    n = c.get_number_of_cycles()
    print(f"number of cycles: {n}")

    cycle = c.get_cap(1, method="forth-and-forth")

    fig_2, (ax4, ax5, ax6) = plt.subplots(1, 3)
    # cycle.plot(x="capacity", y="voltage", ax=ax4)
    s = c.get_step_numbers()
    t = c.sget_timestamp(1, s[1])
    v = c.sget_voltage(1, s[1])
    steps = c.sget_step_numbers(1, s[1])

    print("step numbers:")
    print(s)
    print("sget step numbers:")
    print(steps)
    print("\ntesttime:")
    print(t)
    print("\nvoltage")
    print(v)

    ax5.plot(t, v, label="voltage")
    ax6.plot(t, steps, label="steps")

    fig_3, (ax7, ax8) = plt.subplots(2, sharex=True)
    raw.plot(x="test_time", y="voltage", ax=ax7, title="voltage")
    raw.plot(x="test_time", y="step_index", ax=ax8, title="step index")

    plt.legend()
    plt.show()

    outfile = out / "test_out"
    c.save(outfile)


def check_loader_from_outside_with_get2():
    import pathlib

    import matplotlib.pyplot as plt

    import cellpy
    from cellpy.parameters.internal_settings import headers_normal

    keep = [
        headers_normal.data_point_txt,
        headers_normal.test_time_txt,
        headers_normal.step_time_txt,
        headers_normal.step_index_txt,
        headers_normal.cycle_index_txt,
        headers_normal.current_txt,
        headers_normal.voltage_txt,
        headers_normal.ref_voltage_txt,
        headers_normal.charge_capacity_txt,
        headers_normal.discharge_capacity_txt,
        headers_normal.internal_resistance_txt,
        # "ir_pct_change"
    ]

    # zero: auto
    # one:
    # two: KIT
    # three | WMB_SIMBA: WMG new
    INSTRUMENT = "maccor_txt"
    # MODEL = "WMG_SIMBA"
    # MODEL = "two"
    # MODEL = None
    # MODEL = "KIT_SIMBA"
    MODEL = "S4000-WMG"
    # FILENAME = "1044-CT-MaccorExport.txt"  # WMG_SIMBA
    # FILENAME = "01_UBham_M50_Validation_0deg_01.txt"  # WMG_SIMBA and NONE
    # FILENAME = "KIT-Full-cell-PW-HC-CT-cell002.txt"
    # FILENAME = "KIT-Full-cell-PW-HC-CT-cell016.txt"
    # FILENAME = "KIT-Full-cell-PW-HC-CT-cell013.txt"  # Two
    # FILENAME = "KIT-Full-cell-PW-HC-CT-cell009.txt"  # FAILS in convert_date_time_to_datetime!
    # FILENAME = "SIM-A7-1039 - 073.txt"
    # FILENAME = "SIM-A7-1047-ET - 079.txt"
    FILENAME = "1039_Data from File.txt"
    # FILENAME = "1044_Data from File.txt"
    # FILENAME = "1475_Data from File.txt"

    DATADIR = r"C:\scripting\cellpy_dev_resources\2021_leafs_data\Data from File"

    pd.options.display.max_columns = 100
    datadir = pathlib.Path(DATADIR)
    name = datadir / FILENAME
    out = pathlib.Path(r"C:\scripting\trash")
    print(f"File exists? {name.is_file()}")
    if not name.is_file():
        print(f"could not find {name} ")
        return

    c = cellpy.get(
        filename=name, instrument=INSTRUMENT, model=MODEL, mass=1.0, auto_summary=False
    )
    print(f"loaded the file - now lets see what we got")
    raw = c.cell.raw
    raw.to_clipboard()
    print(raw.head())
    c.make_step_table()

    steps = c.cell.steps
    summary = c.cell.summary

    raw.to_csv(out / "raw.csv", sep=";")
    steps.to_csv(out / "steps.csv", sep=";")
    summary.to_csv(out / "summary.csv", sep=";")

    fig_1, (ax1, ax2, ax3, ax4) = plt.subplots(
        4,
        1,
        figsize=(6, 10),
        constrained_layout=True,
        sharex=True,
    )
    raw.plot(x="test_time", y="voltage", ax=ax1, xlabel="")
    raw.plot(x="test_time", y="current", ax=ax2, xlabel="")
    raw.plot(
        x="test_time", y=["charge_capacity", "discharge_capacity"], ax=ax3, xlabel=""
    )
    raw.plot(x="test_time", y="cycle_index", ax=ax4)
    fig_1.suptitle(f"{name.name}", fontsize=16)

    n = c.get_number_of_cycles()
    print(f"Number of cycles: {n}")

    plt.legend()
    plt.show()

    outfile = out / "test_out"
    c.save(outfile)


if __name__ == "__main__":
    # check_dev_loader2(model="two")
    # check_loader(number=2, model="two")
    check_loader_from_outside_with_get2()
